"""induction machine (electrical circuit model)

"""
import numpy as np
import scipy.interpolate as ip
import scipy.optimize as so
import logging
import pathlib
from .utils import skin_resistance, skin_leakage_inductance, wdg_leakage_inductances
import femagtools.windings
import femagtools.parstudy
import json
import copy
import warnings
import time

EPS = 1e-13


def log_interp1d(x, y, kind='cubic'):
    """logarithmic interpolation function, including extrapolation

    due to the nature of logarithmic functions, it is not possible to have
    a supporting point at 0!

    Arg:
      x, y (arraylike): supporting points
      kind (str): kind of underlying interpolation, default = 'cubic'

    """
    xx = np.asarray(x)
    yy = np.asarray(y)
    # TODO: check for x=0, dimensions, etc.
    logx = np.log10(xx)
    logy = np.log10(yy)
    lin_interp = ip.interp1d(logx, logy, kind=kind, fill_value="extrapolate")
    def log_interp(zz): return np.power(10.0, lin_interp(np.log10(zz)))
    # TODO check replace interp1d
    #lin_interp = ip.make_interp_spline(logx, logy, bc_type='not-a-knot')
    #def log_interp(zz): return np.power(10.0, lin_interp(np.log10(zz), nu=1))
    return log_interp


eecdefaults = dict(
    zeta1=0.2,
    zeta2=2.4,
    gam=0,  # gam=0.7, 0..1
    kh=1,  # kh=2, number of vertical conductors in slot
    tcu1=20,
    kpfe=1,
    rotor_mass=0,
    kfric_b=1,
    pl2v=0.5,
    tcu2=20)

logger = logging.getLogger('im')
logging.captureWarnings(True)


def ring_leakage_inductance(machine):
    """returns the ring leakage inductance
    ref: Design of Rotating Electrical Machines
    Juha Pyrhönen, Tapani Jokinen, Valeria Hrabovcova
    (Ed. 2008)  page 249
    """
    mue0 = 4*np.pi*1e-7

    Qr = machine['rotor']['num_slots']
    m = machine['windings']['num_phases']
    p = machine['poles']//2
    lbar = machine['lfe']
    ls = lbar
    nue = 0.36 if p == 1 else 0.18
    da1 = machine['bore_diam']
    ag = machine['airgap']
    slotmodel = [k for k in machine['rotor'] if isinstance(
        machine['rotor'][k], dict)][-1]

    Dr = (da1 - 2*ag -
          machine['rotor'][slotmodel].get('slot_height', 0) -
          machine['rotor'][slotmodel].get('slot_h1', 0))
    return mue0*Qr/m/p**2/3*((lbar-ls) + nue*np.pi*Dr/2/p)


def slot_opening_factor(n, p, bs, D, Q):
    k = 1 - bs/(D*np.pi/Q)
    return np.sin(n*p*np.pi/Q*(1-k))/(n*p*np.pi/Q*(1-k))


class Component:
    def __init__(self, arg):
        """initialize this object either from a json file
        or from a set of parameters"""
        parameters = arg
        if type(arg) == str:
            file = open(arg)
            parameters = json.load(file)
            file.close()
        for k in parameters.keys():
            setattr(self, k, parameters[k])

    def __str__(self):
        "return string format of this object"
        return repr(self.__dict__)

    def __repr__(self):
        "representation of this object"
        return self.__str__()


class InductionMachine(Component):
    """basic induction machine model with T equivalent circuit"""

    def __init__(self, parameters):
        self.kth1 = self.kth2 = 0.0039  # temp coefficient of stator wdg and rotor bar
        Component.__init__(self, parameters)
        if 'f1type' in parameters:
            self.f1ref = self.f1type
        if 'u1type' in parameters:
            self.u1ref = self.u1type
        if hasattr(self, 'f1ref'):
            self.wref = 2*np.pi*self.f1ref
            if hasattr(self, 'u1ref'):
                self.psiref = self.u1ref/self.wref
        if 'lh' in parameters:
            def imag(self, psi):
                """magnetizing current"""
                return psi/self.lh
            self._imag = imag
        elif 'iml' in parameters:
            def imag(psi):
                return (self.iml * np.abs(psi)/self.psiref +
                        self.ims*np.power(np.abs(psi)/self.psiref, self.mexp))
            self._imag = imag
        elif 'im' in parameters:
            self._imag = log_interp1d(self.psi, self.im)
            self.psi = self.psiref
        if 'pfe' in parameters:
            self.rh = self.m*(self.psiref*self.wref)**2/self.pfe
        for k in eecdefaults.keys():
            if not hasattr(self, k):
                setattr(self, k, eecdefaults[k])
        try:
            self.tfric = self.kfric_b*self.rotor_mass*30e-3/np.pi
            # formula from
            # PERMANENT MAGNET MOTOR TECHNOLOGY: DESIGN AND APPLICATIONS
            # Jacek Gieras
            #
            # plfric = kfric_b m_r n 1e-3 W
            # -- kfric_b : 1..3 W/kg/rpm
            # -- m_r: rotor mass in kg
            # -- n: rotor speed in rpm
        except AttributeError:
            self.tfric = 0

        self.skin_resistance = [None, None]
        # here you can set user defined functions for calculating the skin-resistance,
        # according to the current frequency w. First function in list is for stator, second for rotor.
        # If None, the femagtools intern default implementation is used.
        # User defined functions need to have the following arguments:
        # - r0: (float) dc-resistance at 20°C
        # - w: (float)  current frequency in rad (2*pi*f)
        # - tcu: (float) conductor temperature in deg Celsius
        # - kth: (float) temperature coefficient (Default = 0.0039, Cu)

    def pfric(self, n):
        """friction and windage losses"""
        return 2*np.pi*n*self.tfric


    def imag(self, psi):
        """magnetizing current"""
        if np.isscalar(psi):
            return float(self._imag(psi))
        return self._imag(psi)

    def rfe(self, w, psi):
        """equivalent resistance for iron losses"""
        try:
            if np.isclose(w, 0):
                return 0
            else:
                return self.m*(w*psi)**2 / self.plfe1(w, psi)
        except AttributeError:
            pass
        return self.rh

    def plfe1(self, w1, psi):
        return ((self.fec + self.fee *
                 np.power(np.abs(psi)/self.psiref, self.fexp)) *
                (0.75 + 0.25*(np.abs(w1)/self.wref)) *
                (np.abs(w1)/self.wref) * np.power((np.abs(psi)/self.psiref), 2.0))

    def lrot(self, w):
        """rotor leakage inductance"""
        return skin_leakage_inductance(self.lsigma2, w, self.tcu2,
                                       self.zeta2, 1, self.pl2v)

    def lstat(self, w):
        """stator leakage inductance"""
        # we add end-winding inductance here
        le_value = getattr(self, 'le', 0)
        return self.lsigma1 + le_value

    def rrot(self, w):
        """rotor resistance"""
        sr = self.skin_resistance[1]
        if sr is not None:
            return sr(self.r2, w, self.tcu2, kth=self.kth2)
        else:
            return skin_resistance(self.r2, w, self.tcu2, self.zeta2,
                                   0.0, 1, kth=self.kth2)

    def rstat(self, w):
        """stator resistance"""
        sr = self.skin_resistance[0]
        if sr is not None:
            return sr(self.r1, w, self.tcu1, kth=self.kth1)
        else:
            return skin_resistance(self.r1, w, self.tcu1, self.zeta1,
                                   self.gam, self.kh, kth=self.kth1)
        # return self.r1

    def sigma(self, w, psi):
        """leakage coefficient"""
        lh = psi / self.imag(psi)
        return (1 - lh**2 /
                ((lh+self.lstat(w))*(lh+self.lrot(0))))

    def u1(self, w1, psi, wm):
        """stator voltage"""
        istat = self.i1(w1, psi, wm)
        z1 = (self.rstat(w1) + w1*self.lstat(w1)*1j)
        return w1*psi*1j + istat*z1

    def i1(self, w1, psi, wm):
        """stator current"""
        imag = self.imag(psi)+0j
        if abs(w1) > 0:
            imag += w1*psi/self.rfe(w1, psi)*1j
        return self.i2(w1, psi, wm) + imag

    def i2(self, w1, psi, wm):
        """rotor current"""
        w2 = w1 - self.p * wm
        if abs(w2) > EPS:
            z2 = (self.rrot(w2) + w2*self.lrot(w2)*1j)
            return w2*psi*1j/z2
        return 0

    def w1torque(self, w1, u1max, psi, wm):
        """calculate motor torque"""
        # check stator voltage
        u1 = self.u1(w1, psi, wm)
        self.psi = psi
        if abs(u1) > u1max:  # must adjust flux
            self.psi = so.fsolve(
                lambda psix: u1max - abs(self.u1(w1, psix, wm)),
                psi)[0]
        return self.torque(w1, self.psi, wm)

    def w1tmech(self, w1, u1max, psi, wm):
        """calculate motor shaft torque"""
        # check stator voltage
        u1 = self.u1(w1, psi, wm)
        self.psi = psi
        if abs(u1) > u1max:  # must adjust flux
            self.psi = so.fsolve(
                lambda psix: u1max - abs(self.u1(w1, psix, wm)),
                psi)[0]
        return self.tmech(w1, self.psi, wm)

    def pullouttorque(self, w1, u1):
        """pull out torque"""
        sk = self.sk(w1, u1/w1)
        w2 = sk*w1
        r2 = self.rrot(w2)
        psi = u1/w1
        lh = psi/self.imag(psi)
        x1 = w1*(lh+self.lstat(w1))
        x2 = w1*(lh+self.lrot(w2))
        r1 = self.rstat(w1)
        sigma = self.sigma(w1, u1/w1)
        return self.m*self.p * u1**2/w1 * (1 - sigma) / \
            ((r1**2 + x1**2)*r2/(sk * x1 * x2) +
             sk * x2*(r2**2 + sigma**2 * x1**2)/(r2 * x1) + 2*r1*(1-sigma))

    def sk(self, w1, psi):
        """pullout slip"""
        r2 = self.rrot(0.)
        lh = psi/self.imag(psi)
        x1 = w1*(lh+self.lstat(w1))
        x2 = w1*(lh+self.lrot(0.))
        r1 = self.rstat(w1)
        return r2/x2*np.sqrt((r1**2+x1**2)/(
            self.sigma(w1, psi)**2*x1**2 + r1**2))

    def torque(self, w1, psi, wm):
        """electric torque (in airgap)"""
        w2 = w1-self.p*wm
        if np.isclose(w2, 0):
            return 0.
        r2 = self.rrot(w2)
        i2 = self.i2(w1, psi, wm)
        return self.m*self.p/w2*r2*(i2*i2.conjugate()).real

    def tmech(self, w1, psi, wm):
        """shaft torque"""
        u1 = self.u1(w1, psi, wm)
        return self.torque(w1, psi, wm) - self.tfric

    def torqueu(self, w1, u1max, wm):
        """electric torque (in airgap)"""
        if np.isclose(w1, self.p*wm):
            return 0.
        # find psi
        psi = u1max/w1
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.psi = so.fsolve(
                lambda psi: u1max - abs(self.u1(w1, psi, wm)),
                psi)[0]
        return self.torque(w1, self.psi, wm)

    def tmechu(self, w1, u1max, wm):
        """shaft torque"""
        if np.isclose(w1, self.p*wm):
            return 0.
        # find psi
        psi = u1max/w1
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.psi = so.fsolve(
                lambda psi: u1max - abs(self.u1(w1, psi, wm)),
                psi)[0]
        return self.torque(w1, self.psi, wm)

    def w1(self, u1max, psi, tload, wm, with_tmech=True):
        """calculate stator frequency with given torque and speed"""
        wsync = max(wm*self.p, 0.001)
        # sk = self.rrot(0.)/(wsync*(self.lstat(wsync) + self.lrot(0.0)))
        if tload < 0:
            b = 0.999*wsync
        else:
            b = 1.001*wsync
        if with_tmech:
            logger.debug("wm %s tload %s w1tmech %s",
                         wm, tload, self.w1tmech(b, u1max, psi, wm))
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return so.fsolve(
                    lambda w1: self.w1tmech(w1, u1max, psi, wm) - tload, b)[0]

        logger.debug("wm %s tload %s w1torque %s",
                     wm, tload, self.w1tmech(b, u1max, psi, wm))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return so.fsolve(
                lambda w1: self.w1torque(w1, u1max, psi, wm) - tload, b)[0]

    def wmfweak(self, u1max, psi, torque, with_tmech=True):
        """return lower speed limit of field weakening range"""
        wm0 = u1max/psi/self.p
        return so.fsolve(
            lambda wm: u1max - np.abs(self.u1(
                self.w1(u1max, psi, torque, wm, with_tmech), psi, wm)),
            wm0)[0]

    def torque_chart(self, smin=-0.1, smax=0.1, nsamples=100, with_tmech=True):
        """
        calculate torque(s) curve

        :param smin: min slip
        :param smax: max slip
        :param with_tmech: use friction and windage losses (shaft torque)
        :return: dict with slip and torque lists
        """

        s = np.linspace(smin, smax, nsamples)
        f1 = self.f1ref
        u1 = self.u1ref
        w1 = 2 * np.pi * f1
        wm = (1-s) * w1 / self.p
        r = {}
        r['s'] = list(s)
        if with_tmech:
            r['T'] = [self.tmechu(w1, u1, wx) for wx in wm]
        else:
            r['T'] = [self.torqueu(w1, u1, wx) for wx in wm]
        return r

    def operating_point(self, T, n, u1max, Tfric=None, **kwargs):
        """
        calculate single operating point.

        return dict with values for
            s -- (float) slip at this op
            f1 -- (float) stator frequency in Hz
            u1 -- (float) stator phase voltage in V rms
            i1 -- (float) stator phase current in A rms
            i2 -- (float) rotor bar current in A rms
            p1 -- (float) electrical power in W
            pmech -- (float) mechanical power in W
            plcu1 -- (float) stator copper losses in W
            plcu2 -- (float) rotor copper losses in W
            plfric -- (float) friction losses in W, according to Tfric
            losses -- (float) total losses in W
            cosphi -- (float) power factor for this op
            eta -- (float) efficiency for this op
            T -- (float) torque for this op in Nm, copy of argument T
            Tfric -- (float) friction torque in Nm
            n -- (float) speed for this op in 1/s, copy of argument n
            tcu1/2 -- (float) temperature of statur/rotor in deg. C

        Keyword arguments:
            T -- (float) the output torque at the shaft in Nm
            n -- (float) the speed of the machine in 1/s
            u1max -- (float) the maximum phase voltage in V rms
            Tfric -- (float, optional) the friction torque to consider in Nm. If Tfric is None, the friction torque is calculated according to the rotor-mass and the friction factor of the machine (kfric_b). Friction is written to the result dict!
            with_tmech -- (optional) use friction and windage losses
        """

        r = {}  # result dit

        if Tfric:
            tfric = Tfric
        else:
            tfric = self.tfric
            # TODO: make frictiontorque speed depended?

        with_tmech = kwargs.get('with_tmech', True)
        if with_tmech:
            tq = T + tfric
        else:
            tq = T
        wm = 2*np.pi*n
        w1 = self.w1(u1max, self.psiref, tq, wm, with_tmech)
        s = (w1 - self.p*wm) / w1
        r['s'] = float(s)
        r['f1'] = float(w1/2/np.pi)
        u1 = self.u1(w1, self.psi, wm)
        r['u1'] = float(np.abs(u1))
        i1 = self.i1(w1, self.psi, wm)
        r['i1'] = float(np.abs(i1))
        r['cosphi'] = float(np.cos(np.angle(u1) - np.angle(i1)))
        r['plfe1'] = self.plfe1(w1, self.psi)
        i2 = self.i2(w1, self.psi, wm)
        r['i2'] = float(np.abs(i2))
        r['plcu1'] = float(self.m * np.abs(i1) ** 2 * self.rstat(w1))
        r['plcu2'] = float(self.m * np.abs(i2) ** 2 *
                           self.rrot(w1 - self.p * wm))
        r['plfw'] = float(self.pfric(n))
        pmech = 2 * np.pi * n * tq
        r['pmech'] = float(pmech)
        pltotal = r['plfe1'] + r['plfw'] + r['plcu1'] + r['plcu2']
        r['losses'] = float(pltotal)
        p1 = pmech + pltotal
        r['p1'] = float(p1)
        if np.abs(pmech) < 1e-12:
            eta = 0  # power to low for eta calculation
        elif p1 > pmech:
            eta = pmech/p1  # motor
        else:
            eta = p1/pmech  # generator
        r['eta'] = float(eta)
        r['T'] = float(T)
        r['Tfric'] = float(tfric)
        r['n'] = float(n)
        r['tcu1'] = float(self.tcu1)
        r['tcu2'] = float(self.tcu2)

        return r

    def characteristics(self, T, n, u1max, nsamples=70, kpo=0.9, **kwargs):
        """calculate torque speed characteristics.
        return dict with list values of
        id, iq, n, T, ud, uq, u1, i1, s, pullout slip
        beta, gamma, phi, cosphi, pmech, n_type

        Keyword arguments:
        T -- (float) the maximum torque in Nm
        n -- (float) the maximum speed in 1/s
        u1max -- (float) the maximum phase voltage in V rms
        nsamples -- (optional) number of speed samples
        with_tmech -- (optional) use friction and windage losses

        """
        with_tmech = kwargs.get('with_tmech', True)

        if np.isscalar(T):
            wmType = self.wmfweak(u1max, self.psiref, T, with_tmech)
            logger.info(wmType)
            pmmax = wmType*T
            wmPullout = so.fsolve(
                lambda wx: (kpo*self.pullouttorque(self.p *
                            wx, u1max) - abs(pmmax/wx)),
                wmType)[0]
            wmtab0 = np.linspace(wmType, 3*wmPullout)
            for wm, tq in zip(wmtab0, [pmmax/wx for wx in wmtab0]):
                logger.debug("u1 %g psi %g tq %g wm %g",
                            u1max, self.psiref, tq, wm)
                try:
                    w1 = self.w1(u1max, self.psiref, tq, wm, with_tmech)
                except ValueError:
                    wmPullout = wm
                    break
            wmMax = max(1.5*wmPullout, 3*abs(pmmax/T))
            if n:
                wmMax = 2*np.pi*n
            if wmMax > wmPullout:
                wmtab0 = np.linspace(wmPullout, wmMax)
                for wm, tq in zip(wmtab0, [pmmax/wx**2
                                        for wx in wmtab0]):
                    logger.debug("u1 %g psi %g tq %g wm %g",
                                u1max, self.psiref, tq, wm)
                    try:
                        w1 = self.w1(u1max, self.psiref, tq, wm, with_tmech)
                    except ValueError:
                        wmMax = wm
                        break

            logger.info("wmtype %f wpo %f wmmax %f", wmType, wmPullout, wmMax)

            if wmType < wmPullout < wmMax:
                wmrange = sorted([0, wmType, wmPullout, wmMax])
            elif wmMax < wmType:
                wmrange = sorted([0, wmMax])
            elif wmType<0 or wmPullout<0 or wmMax<0:
                wmrange = sorted([1, abs(min(wmMax,wmType, wmPullout)), abs(max(wmMax,wmType, wmPullout))])
            else:
                wmrange = sorted([0, wmType, wmMax])
            logger.info("Speed range %s", wmrange)
            wmlin = []
            dw = 0
            for i, nx in enumerate([round(nsamples*(w1-w0)/wmMax)
                                    for w1, w0 in zip(wmrange[1:],
                                                    wmrange)]):
                if nx == 1:
                    nx = 2
                if nx > 0:
                    lw = np.linspace(wmrange[i]+dw, wmrange[i+1], nx)
                    dw = lw[-1] - lw[-2]
                    wmlin.append(lw)
            if len(wmlin) > 1:
                wmtab = np.concatenate(wmlin)
            else:
                wmtab = wmlin[1:]

            def tload2(wm):
                if wm < wmType and wm < wmPullout:
                    return T
                if wm < wmPullout:
                    if pmmax < 0:
                        return max(T, pmmax/wm)
                    return min(T, pmmax/wm)
                if pmmax < 0:
                    return max(wmPullout*pmmax/wm**2, T)
                return min(wmPullout*pmmax/wm**2, T)

            T = [tload2(wx) for wx in wmtab]
        else:
            wmtab = n
        r = dict(u1=[], i1=[], T=[], cosphi=[], n=[], s=[], sk=[],
                    plfe1=[], plcu1=[], plcu2=[], f1=[])
        tfric = self.tfric
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for wm, tq in zip(wmtab, T):
                logger.debug("u1 %g psi %g tq %g wm %g",
                             u1max, self.psiref, tq, wm)
    #            try:
                w1 = self.w1(u1max, self.psiref, tq, wm, with_tmech)
                r['f1'].append(w1/np.pi/2)
                u1 = self.u1(w1, self.psi, wm)
                r['u1'].append(np.abs(u1))
                i1 = self.i1(w1, self.psi, wm)
                r['i1'].append(np.abs(i1))
                r['cosphi'].append(np.cos(np.angle(u1) - np.angle(i1)))
                r['plfe1'].append(self.plfe1(w1, self.psi))
                i2 = self.i2(w1, self.psi, wm)
                r['plcu1'].append(self.m*np.abs(i1)**2*self.rstat(w1))
                r['plcu2'].append(self.m*np.abs(i2)**2*self.rrot(w1-self.p*wm))
                if with_tmech:
                    r['T'].append(tq)
                else:
                    r['T'].append(tq - tfric)
                r['n'].append(wm/2/np.pi)
                r['s'].append(float((w1 - self.p * wm) / w1))
                r['sk'].append(self.sk(w1, np.abs(u1)/w1))
                # add n_type to result dict
                if np.isscalar(T):
                    r['n_type'] = wmType/2/np.pi
                #            except ValueError as ex:
                #                break
        r['plfw'] = [self.pfric(n) for n in r['n']]
        r['pmech'] = [2*np.pi*n*tq for n, tq in zip(r['n'], r['T'])]
        pmech = np.array(r['pmech'])
        pltotal = (np.array(r['plfe1']) + np.array(r['plfw']) +
                   np.array(r['plcu1']) + np.array(r['plcu2']))
        r['losses'] = pltotal.tolist()

        if pmech.any():
            p1 = pmech + pltotal
            if np.abs(pmech[0]) < 1e-12:
                r['eta'] = [0]
                i = 1
            else:
                r['eta'] = []
                i = 0
            if np.all(abs(p1[i:]) > abs(pmech[i:])):
                r['eta'] += (pmech[i:]/(p1[i:])).tolist()
            else:
                r['eta'] += (p1[i:]/pmech[i:]).tolist()

        return r


def parident(workdir, engine, f1, u1, wdgcon,
             machine, magnetizingCurves, condMat=[], **kwargs):
    """return dict of parameters of equivalent circuit for induction machines

    arguments:
    workdir --directory for intermediate files
    engine -- calculation driver (multiproc, docker, condor)
    f1 -- line voltage frequency [Hz]
    u1 -- line voltage (line-to-line) [V]
    wdgcon -- winding connection  ('open', 'wye', 'star', 'delta')
    machine -- dict() with machine parameters
    magnetizingCurves -- list of dict() with BH curves (or directory with MC/MCV files)
    condMat -- list of dict() with conductor material properties

    optional:
    num_steps: (int) number of current steps for the no-load calculation (default 5)
    logspace: (bool) uses log distributed current samples if true linear otherwise (default true)
    i_max_fact: (float) factor for maximum current to calculate no_load flux (default=2.5)
    templatedirs: (list of str) names of directories to search for templates
    """
    cmd = kwargs.get("cmd", None)
    CON = {'open': 0, 'wye': 1, 'star': 1, 'delta': 2}
    p = machine['poles']//2
    slip = 1e-2
    u1ph = u1
    w1 = 2*np.pi*f1
    wdgk = 'windings' if 'windings' in machine else 'winding'
    if CON[wdgcon] == 1:
        u1ph /= np.sqrt(3)
    wdg = femagtools.windings.Winding(
        {'Q': machine['stator']['num_slots'],
         'm': machine[wdgk]['num_phases'],
         'p': machine['poles']//2,
         'l': machine[wdgk]['num_layers'],
         'yd': machine[wdgk]['coil_span']})
    L1 = wdg.inductance(
        nwires=machine[wdgk]['num_wires'],
        g=machine[wdgk]['num_par_wdgs'],
        da1=machine['bore_diam'],
        lfe=machine['lfe'],
        ag=machine['airgap'])
    i1max_fact = kwargs.get('i_max_fact', 2.5)
    i1max = i1max_fact*u1ph/w1/L1
    i1min = u1ph/w1/L1/5
    num_steps = kwargs.get('num_steps', 5)
    if kwargs.get('logspace', False):
        b = (i1min-i1max)/np.log(i1min/i1max)
        a = i1max/b
        i1tab = [b*(a+np.log(x))
                 for x in np.linspace(i1min/i1max, 1,
                                      num_steps)]
    else:
        i1tab = np.linspace(i1min, i1max, num_steps).tolist()

    m = copy.deepcopy(machine)
    Q2 = m['rotor']['num_slots']
    noloadsim = dict(
        calculationMode="noloadflux-rot",
        curvec=i1tab,
        num_par_wdgs=machine[wdgk].get('num_par_wdgs', 1),
        Q2=Q2)

    da1 = m['bore_diam']
    ag = m['airgap']
    # do not create airgap nodechains automatically
    # we prefer an integer number of elements per rotor slot
    nper = np.lcm(wdg.Q, Q2)/Q2
    m['num_agnodes'] = Q2*np.ceil(2*(da1-ag)*np.pi/Q2/ag/nper)*nper
    try:
        m[wdgk].pop('resistance')
    except KeyError:
        pass
    
    parstudy = femagtools.parstudy.ParameterStudy(
        workdir, condMat=condMat,
        magnetizingCurves=magnetizingCurves, cmd=cmd)
    
    builder = femagtools.fsl.Builder(kwargs.get('templatedirs', []))
    model = femagtools.model.MachineModel(m)
    # modelfiles = parstudy.setup_model(builder, model)
    extra_result_files = [f'noloadbag-{i+1}.dat'
                          for i in range(num_steps)] + [
        'psi-rot-mag.dat']

    # set AC simulation
    slotmodel = [k for k in machine['rotor'] if isinstance(
        machine['rotor'][k], dict)][-1]
    Dr = (da1 - 2*ag -
          machine['rotor'][slotmodel].get('slot_height', 0) -
          machine['rotor'][slotmodel].get('slot_h1', 0))
    bar_len = machine['lfe']+np.pi*Dr/Q2/np.sin(np.pi*p/Q2)

    ecsim = dict(
        calculationMode="ec-rotorbar",
        bar_len=bar_len,
        wind_temp=20,
        bar_temp=20,
        f1=f1)

    rotorbar = copy.deepcopy(machine)
    # remove stator slot model and winding
    for k in rotorbar['stator']:
        if isinstance(rotorbar['stator'][k], dict):
            d = rotorbar['stator'].pop(k)
            break
    d = rotorbar.pop(wdgk)
    if 'shaft_diam' in rotorbar:
        rotorbar['inner_diam'] = rotorbar.pop('shaft_diam')
    # use one rotor slot only
    rotorbar['stator']['statorRing'] = {'num_slots': Q2}
    for k in rotorbar['rotor']:
        if isinstance(rotorbar['rotor'][k], dict):
            rotorbar['rotor'][k]['num_slots_gen'] = 1
            rotorbar['rotor'][k]['zeroangle'] = 90-180/Q2
            break

    loadsim = dict(  # not used
        calculationMode="asyn_motor",
        bar_len=bar_len,
        wind_temp=20,
        bar_temp=20,
        speed=(1-slip)*f1/p,
        f1=f1,
        num_par_wdgs=machine[wdgk].get('num_par_wdgs', 1),
        wdgcon=CON[wdgcon],  # 0:open, 1:star, 2:delta
        u1=u1ph)  # phase voltage

    # prepare calculation
    job = engine.create_job(workdir)
    pmod = model.poles * model.stator['num_slots_gen'] / model.stator['num_slots']
    task = job.add_task(_eval_noloadrot(pmod), extra_result_files)
    logger.debug("Task %s noload workdir %s result files %s",
                 task.id, task.directory, task.extra_result_files)
    # create model
    for mc in parstudy.femag.copy_magnetizing_curves(
            model,
            dir=task.directory):
        task.add_file(mc)

    task.add_file(
        'femag.fsl',
        builder.create_model(model, condMat=parstudy.femag.condMat) +
        builder.create_analysis(noloadsim) +
        ['save_model("close")'])

    # ec simulation
    barmodel = femagtools.model.MachineModel(rotorbar)
    extra_result_files = ['bar.dat']
    r = (da1-ag)/2
    #task = job.add_task(_eval_ecsim())
    task = job.add_task(_eval_ecsim(), extra_result_files)
    logger.debug("Task %s rotobar workdir %s result files %s",
                 task.id, task.directory, task.extra_result_files)
    task.set_stateofproblem('mag_dynamic')
    for mc in parstudy.femag.copy_magnetizing_curves(
            barmodel,
            dir=task.directory,
            recsin='cur'):  # todo cur
        task.add_file(mc)
    task.add_file(
        'femag.fsl',
        builder.create_model(barmodel,
                             condMat=parstudy.femag.condMat) +
        builder.create_analysis(ecsim) +
        ['save_model("close")'])

    # AC simulation
    actask = job.add_task(result_files=['end_wind_leak.dat'])
    logger.debug("Task %s loadsim workdir %s result files %s",
                 task.id, task.directory, task.extra_result_files)
    actask.set_stateofproblem('mag_dynamic')
    for mc in parstudy.femag.copy_magnetizing_curves(
            model,
            dir=actask.directory,
            recsin='flux'):
        actask.add_file(mc)
    actask.add_file(
        'femag.fsl',
        builder.create_model(model, condMat=parstudy.femag.condMat) +
        builder.create_analysis(loadsim) +
        ['save_model("close")'])
    # start FE simulations
    tstart = time.time()
    status = engine.submit()
    logger.info('Started %s', status)
    status = engine.join()
    tend = time.time()
    logger.info("Elapssed time %d s Status %s",
                (tend-tstart), status)
    if any([x != 'C' for x in status]):
        raise ValueError("AC simulation failed")
    # collect results
    results = [t.get_results() for t in job.tasks]

    i1_0 = results[0]['i1_0'].tolist()
    psi1_0 = results[0]['psi1_0'].tolist()
    # amplitude of flux density in airgap
    bamp = results[0]['Bamp']
    taup = np.pi*(da1-ag)/2/p
    lfe = machine['lfe']
    n1 = wdg.turns_per_phase(machine[wdgk]['num_wires'],
                             machine[wdgk]['num_par_wdgs'])
    # main flux per phase at no load in airgap
    kfe = machine['stator'].get('fillfac', 1.0)
    psih = np.mean([[2/np.pi*n1*wdg.kw()*taup*lfe*b/np.sqrt(2)
                   for b in bb]
                    for bb in bamp], axis=1)
    i10tab = [0] + i1_0
    psihtab = [0] + psih.tolist()
    u1ref = u1ph
    psiref = u1ref/w1

  # def inoload(x, iml, ims, mexp):
  #    """return noload current"""
  #    return iml*x/psiref + ims*(x/psiref)**mexp
  # fitp, cov = so.curve_fit(inoload, psihtab, i10tab, (1, 1, 1))
  # iml, ims, mexp = fitp
  # logger.info("iml, ims, mexp %g, %g, %g",
  #            iml, ims, mexp)
  # i1tab.insert(0, 0)
  # psi1_0.insert(0, 0)
  # i1_0.insert(0, 0)

    logger.info("psi1_0 %s", np.mean(psi1_0, axis=1))
    logger.info("psih %s", psih)
    logger.debug("psi1_0-psih %s", np.mean(psi1_0, axis=1)-psih)
    logger.debug("i1tab %s", i1tab)
    logger.debug("u1ref %s", u1ref)
    logger.debug("w1 %s", w1)
    logger.debug("L1 %s", L1)
    try:
        r1 = machine[wdgk]['resistance']
    except KeyError:
        from .utils import wdg_resistance
        slotmodel = [k for k in machine['stator'] if isinstance(
            machine['stator'][k], dict)][-1]
        if slotmodel == 'stator1':
            hs = machine['stator']['stator1']['slot_rf1'] - \
                machine['stator']['stator1']['tip_rh1']
        else:
            hs = machine['stator'][slotmodel].get('slot_height',
                                                  0.33*(machine['outer_diam']-da1))
        n = machine[wdgk]['num_wires']
        if 'dia_wire' in machine[wdgk] and machine[wdgk].get('dia_wire', 1e-3) !=0:
            aw = np.pi*machine[wdgk].get('dia_wire', 1e-3)**2/4
        else:  # wire diameter from slot area
            aw = 0.75 * \
                machine[wdgk].get(
                    'cufilfact', 0.45)*np.pi*da1*hs/wdg.Q/2/n
        # TODO: read nc file and get slot area:
        # as = nc.windings[0].subregions[0].area()
        # aw = as/wdg.l/n*fillfac
        # TODO: sigma = 58e6
        # if 'material' in machine[wdgk]:
        #    sigma = condMat[machine[wdgk]]['elconduct']
        if 'material' in m['windings']:
            sigma = m['windings']['cuconduct']
        g = loadsim['num_par_wdgs']
        r1 = wdg_resistance(
            wdg, n, g, aw, da1, hs, lfe, sigma=sigma)

    # psi1 = ip.interp1d(i1_0, np.mean(psi1_0, axis=1),
    #                    kind='quadratic')
    psi1 = log_interp1d(i1_0, np.mean(psi1_0, axis=1))
    imref = so.fsolve(
        lambda imx: u1ref - np.sqrt((imx*r1)**2 +
                                    (w1*psi1(imx))**2),
        u1ph/w1/L1)[0]
    # psihref = float(ip.interp1d(i1tab, psih,
    #                             kind='quadratic')(imref))
    psihref = float(log_interp1d(i1tab, psih)(imref))
    psi1ref = float(psi1(imref))
  

    lh = psihref/imref
    L1 = psi1ref/imref
    ls1 = L1 - lh*(1+wdg.harmleakcoeff())
    logger.info('L1: %g Lh: %g Ls1: %g kw %g, sigma0: %g',
                L1, lh, ls1, wdg.kw(), wdg.harmleakcoeff())
    ü = 3*4*(wdg.kw()*n1)**2/Q2
    r2 = results[1]['r2']*ü
    ls2 = results[1]['ls2']*ü + ring_leakage_inductance(machine)*ü
    zeta2 = results[1]['zeta2']
    pl2v = results[1]['pl2v']

    # end winding leakage
    basedir = pathlib.Path(actask.directory)
    leakfile = basedir / 'end_wind_leak.dat'
    try:
        leakages = [float(x)
                    for x in leakfile.read_text().split()]
        ls1 += leakages[1]  # TODO: np.linalg.norm(leakages[1:])
        logger.info("L1ew %s", leakages[1])
    except:
        logger.warning("No end winding leakage")
        Lu, Lew = wdg_leakage_inductances(machine)
        logger.info("L1ew %s", Lew)
        ls1 += Lew
    p = results[2]['p']
    m = results[2]['num_phases']
    # r2 = results[1]['r2']
    # ls2 = results[1]['ls2']
    pfe = results[2]['pfe1'][0]

    end_ring_section = machine['rotor'].get('Aring', 0)
    bore_diam = machine['bore_diam']
    for key, value in machine['rotor'].items():
        if isinstance(value, dict) and 'slot_height' in value:
            slot_height_rotor = value['slot_height']
            break
    if slot_height_rotor is not None:
        rayon = (bore_diam / 2) - slot_height_rotor/2
        end_ring_length = 2 * np.pi * rayon
    else:
        end_ring_length = 0

    end_ring_volume = end_ring_section*end_ring_length

    for material_info in condMat:
        if material_info.get('name') == machine.get('rotor', {}).get('material'):
            spmaweight_rotor = material_info.get('spmaweight')
            break
        else:
            spmaweight_rotor = 0
    end_ring_mass = end_ring_volume * spmaweight_rotor*1e3

    rotor_mass = sum([results[2].get('conweight', 0),
                      results[2].get('lamweight', 0)])
    logger.info('Conductor mass in rotor = %s',results[2].get('conweight', 0))
    logger.info('Lamination mass in rotor = %s',results[2].get('lamweight', 0))
    logger.info('Rotor mass without end-ring = %s',rotor_mass)
    rotor_mass += 2*end_ring_mass
    logger.info('One end-ring mass = %s',end_ring_mass)
    logger.info('Total rotor mass = %s',rotor_mass)

    n = machine[wdgk]['num_wires']
    g = loadsim['num_par_wdgs']
    impars = {
        'p': p, 'm': m,
        'f1ref': f1, 'u1ref': u1ph,
        'rotor_mass': rotor_mass, 'kfric_b': 1,
        'r1': r1, 'r2': r2,
        'lsigma1': ls1,
        'lsigma2': ls2, 'zeta2': zeta2, 'pl2v': pl2v,
        'psiref': psihref, 'wref': w1,
        'fec': pfe, 'fee': 0, 'fexp': 7.0,
        'im': i1tab, 'psi': psih.tolist()}
    try:
        wmat = machine[wdgk]['material']
        impars['kth1'] = parstudy.femag.condMat.find(wmat)['tempcoef']
    except KeyError:
        logger.warning('Missing winding material id for stator')
        for mat in condMat:
            if mat['name'] == 'Cu': #as Copper is used as default (cf r1 and wdg_resistance)
                impars['kth1'] = mat['tempcoef']
    try:
        bmat = machine['rotor']['material']
        impars['kth2'] = parstudy.femag.condMat.find(bmat)['tempcoef']
    except KeyError:
        logger.warning('Missing winding material id for rotor')

    return impars


class _eval_noloaddc():
    """ Result Functor for noloadflux dc calc"""

    def __init__(self, pmod):
        self.pmod = pmod

    def __call__(self, task):
        basedir = pathlib.Path(task.directory)
        psimag = np.loadtxt(basedir/'noloadflux.dat').T
        im, psi = psimag[0:6:2], psimag[12::2]
        d = 0
        a = np.array(
            (np.cos(d), np.cos(d-2*np.pi/3), np.cos(d+2*np.pi/3)))/3*2
        i0 = np.sum(a*im.T, axis=1)/np.sqrt(2)
        psi0 = np.sum(a*psi.T, axis=1)/np.sqrt(2)

        bag = np.loadtxt(basedir / 'noloadbag.dat').T
        Bamp = [b['Bamp'] for b in [
            femagtools.airgap.fft(bag[0], b, self.pmod)
            for b in bag[1:]]]
        return {'i1_0': i0, 'psi1_0': psi0, 'Bamp': Bamp}


class _eval_noloadrot():
    """ Result Functor for noloadflux rot calc"""

    def __init__(self, pmod):
        self.pmod = pmod

    def __call__(self, task):
        basedir = pathlib.Path(task.directory)
        psimag = np.loadtxt(basedir/'psi-rot-mag.dat')
        pos = np.unique(psimag[:, 0])
        ncurs = psimag[:, 0].shape[0]//pos.shape[0]
        ire = psimag[:ncurs, 1:7:2]  # assuming same current for all steps
        psire = np.reshape(psimag[:, 7:12:2], (-1, ncurs, 3))

        i0 = np.linalg.norm(
            femagtools.machine.T(0).dot(ire.T),
            axis=0)/np.sqrt(2)
        

        psi0 = np.array([np.linalg.norm(
            femagtools.machine.T(0).dot(psire[:, k, :].T),
            axis=0)/np.sqrt(2)
            for k in range(ncurs)])
        
        # matrix (i x j x k) of curr, rotor pos, angle
        Bamp = [[femagtools.airgap.fft(bags[:, 0], b, self.pmod)['Bamp']
                 for b in bags.T[1:]]
                for bags in [np.loadtxt(p) for p in sorted(
                    basedir.glob(f"noloadbag-*.dat"),
                    key=lambda path: int(path.stem.rsplit("-", 1)[1]))]]
        return {'i1_0': i0, 'psi1_0': psi0, 'Bamp': Bamp}


class _eval_ecsim():
    """ Result Functor for ec simulation"""

    def __init__(self):
        pass

    def __call__(self, task):
        import lmfit
        from .utils import xiskin, kskinl
        basedir = pathlib.Path(task.directory)
        freqrind = np.loadtxt(basedir/'bar.dat').T
        rbar = freqrind[1]
        lbar = freqrind[2:]
        f = freqrind[0]
        r0 = rbar[0]
        l0 = lbar[0, 0]
        temp = 20

        def barimp(f, zeta, pl2v):
            w = 2*np.pi*f
            r = skin_resistance(r0, w, temp, zeta)
            xi = xiskin(w, temp, zeta)
            lbar = l0*(1 + pl2v*(kskinl(xi, 1)-1))
            return r + 1j*lbar
        model = lmfit.model.Model(barimp)
        params = model.make_params(zeta=1, pl2v=0.5)
        guess = lmfit.models.update_param_vals(params, model.prefix)
        imp = rbar + 1j*lbar[-1]
        rfit = model.fit(imp, params=guess, f=f, verbose=True)
        zeta = rfit.params['zeta'].value
        pl2v = rfit.params['pl2v'].value
        logger.info("ECSIM %s",
                    {'r2': r0, 'ls2': l0, 'zeta2': zeta, 'pl2v': pl2v})
        return {'r2': r0, 'ls2': l0, 'zeta2': zeta, 'pl2v': pl2v}


class _eval_end_wind_leak():
    """ Result Functor for ac simulation"""

    def __init__(self):
        pass

    def __call__(self, task):
        basedir = pathlib.Path(task.directory)
        leakfile = basedir / 'end_wind_leak.dat'
        if leakfile.exists():
            leakages = leakfile.read_text().split()
            return {'lse': np.linalg.norm(leakages[1:])}
        return {'lse': 0}


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import femagtools.plot

    mpars = dict(
        u1ref=2184/np.sqrt(3),
        f1ref=53,
        m=3,
        p=2,
        rh=524.2776,
        # lh=35.88343e-3,
        r1=0.043,
        r2=0.0346,
        lsigma1=1.15e-3,
        lsigma2=0.73361e-3,
        fec=9100,
        fee=0,
        fexp=7.0,
        #pfric=2544.6,
        rexp=2.4,
        iml=79.8286,
        ims=27.5346,
        mexp=6.5)

    im = InductionMachine(mpars)
    torque = 6543
    u1max = 2184/np.sqrt(3)
    r = im.characteristics(torque, 0, u1max)
    # plt.plot(np.array(r['n'])*60, r['cosphi'])
    #    plt.plot(np.array(r['n'])*60, r['plcu'])
    #    plt.plot(np.array(r['n'])*60, r['losses'])
    femagtools.plot.characteristics(r)
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
import femagtools.magnet

magMat = [dict(
    name="M395",
    remanenc=1)
]


def test_findById():
    mag = femagtools.magnet.Magnet(magMat)
    result = mag.find('M395')
    expected = magMat[0]
    assert result == expected



"""plot nc model

"""
# https://stackoverflow.com/questions/42386372/increase-the-speed-of-redrawing-contour-plot-in-matplotlib/42398244#42398244
import numpy as np
import matplotlib.pyplot as plt
import logging


logger = logging.getLogger("femagtools.plot.nc")


DEFAULT_CMAP='viridis'
"""default colormap (see https://matplotlib.org/stable/users/explain/colors/colormaps.html)"""


def spel(isa, superelements=[], with_axis=False, with_wiredir=False, ax=0):
    """plot super elements of I7/ISA7 model
    Args:
      isa: Isa7 object
      superelements: list of super elements (all if empty)
    """
    from matplotlib.patches import Polygon
    if ax == 0:
        ax = plt.gca()
    ax.set_aspect('equal')
    if superelements:
        spels = superelements
    else:
        spels = isa.superelements
    for se in spels:
        ax.add_patch(Polygon([n.xy
                            for nc in se.nodechains
                            for n in nc.nodes],
                            color=isa.color[se.color], lw=0))
        try:
            # draw wire direction
            if se.subregion and with_wiredir:
                if se.subregion.curdir != 0:
                    wkey = se.subregion.winding.key
                    if se.subregion.curdir < 0:
                        label = str(se.subregion.curdir*wkey)
                    else:
                        label = '+'+str(se.subregion.curdir*wkey)

                    xy = np.array([n.xy
                        for nc in se.nodechains
                        for n in nc.nodes])
                    cx, cy = (np.mean(np.unique(xy[:, 0])),
                              np.mean(np.unique(xy[:, -1])))
                    ax.text(cx, cy, label, rotation=np.arctan2(cy, cx)/np.pi*180-90,
                            horizontalalignment='center', verticalalignment='center')
        except:
            pass
    ax.autoscale(enable=True)
    if not with_axis:
        ax.axis('off')


def mesh(isa, with_axis=False, ax=0):
    """plot mesh of I7/ISA7 model
    Args:
      isa: Isa7 object
    """
    from matplotlib.lines import Line2D
    if ax == 0:
        ax = plt.gca()
    ax.set_aspect('equal')
    for el in isa.elements:
        z = np.array([v.xy for v in el.vertices])
        pts = np.vstack((z, z[0])).T
        ax.add_line(Line2D(pts[0], pts[1],
                           color='b', ls='-', lw=0.25))

    # for nc in isa.nodechains:
    #    pts = [list(i) for i in zip(*[(n.x, n.y) for n in nc.nodes])]
    #    ax.add_line(Line2D(pts[0], pts[1], color="b", ls="-", lw=0.25,
    #                       marker=".", ms="2", mec="None"))

    # for nc in isa.nodechains:
    #    if nc.nodemid is not None:
    #        plt.plot(*nc.nodemid.xy, "rx")

    ax.autoscale(enable=True)
    if not with_axis:
        ax.axis('off')


def _contour(ax, title, elements, values, label='',
             cmap=DEFAULT_CMAP, isa=None, alpha=1,
             relminrange=0.05):
    from matplotlib.patches import Polygon
    from matplotlib.collections import PatchCollection
    if ax == 0:
        ax = plt.gca()
    ax.set_aspect('equal')
    if title:
        ax.set_title(title)
    if isa:
        for se in isa.superelements:
            ax.add_patch(Polygon([n.xy
                                  for nc in se.nodechains
                                  for n in nc.nodes],
                                 color='gray', alpha=0.1, lw=0))
    valid_values = np.logical_not(np.isnan(values))
    vertices = [[v.xy for v in e.vertices] for e in elements]
    patches = np.array([Polygon(xy) for xy in vertices])[valid_values]
    p = PatchCollection(patches, match_original=False,
                        cmap=cmap, alpha=alpha)
    z = values[valid_values]
    clim = np.min(z), np.max(z)
    n = np.max(np.abs(clim))
    if not np.isclose(n, 0):
        relrange = (clim[1]-clim[0])/n
        if relrange < relminrange:  # prevent strange color pattern
            c = (clim[1]+clim[0])/2
            p.set_clim([(1-relminrange)*c, (1+relminrange)*c])
    p.set_array(z)
    ax.add_collection(p)
    cb = plt.colorbar(p, shrink=0.9)

    for patch in np.array([Polygon(xy, fc='white', alpha=1.0)
                           for xy in vertices])[np.isnan(values)]:
        ax.add_patch(patch)
    if label:
        cb.set_label(label=label)
    ax.autoscale(enable=True)
    ax.axis('off')


def demag(isa, cmap=DEFAULT_CMAP, relminrange=0.05, ax=0):
    """plot demag of NC/I7/ISA7 model
    Args:
      isa: Isa7/NC object
    """
    emag = [e for e in isa.elements if e.is_magnet()]
    demag = np.array([e.demagnetization(isa.MAGN_TEMPERATURE) for e in emag])
    _contour(ax, f'Demagnetization at {isa.MAGN_TEMPERATURE} °C (max -{np.max(demag):.1f} kA/m)',
             emag, demag, '-H / kA/m', cmap=cmap, isa=isa,
             relminrange=relminrange)
    logger.info("Max demagnetization %f", np.max(demag))


def remanence(isa, cmap=DEFAULT_CMAP, relminrange=0.05, ax=0):
    """plot remanence of NC/I7/ISA7 model
    Args:
      isa: Isa7/NC object
    """
    emag = [e for e in isa.elements if e.is_magnet()]
    rem = np.linalg.norm([e.remanence(isa.MAGN_TEMPERATURE)
                          for e in emag], axis=1)
    _contour(ax, f'Remanence at {isa.MAGN_TEMPERATURE} °C (min {np.min(rem):.1f} T)',
             emag, rem, 'T', cmap=cmap, isa=isa, relminrange=relminrange)
    logger.info("Min remanence %f", np.min(rem))


def demag_pos(isa, pos=-1, icur=-1, ibeta=-1, cmap=DEFAULT_CMAP,
              relminrange=0.05, ax=0):
    """plot demag of NC/I7/ISA7 model at rotor position
    Args:
        isa: Isa7/NC object
        pos: rotor position in degree (maximum h if -1)
        icur: cur amplitude index or last index if -1
        ibeta: beta angle index or last index if -1
        cmap: colormap
        relminrange: rel minimum range for colorbar
    """
    emag = [e for e in isa.elements if e.is_magnet()]
    demag = np.array([isa.demagnetization(e, icur, ibeta)[1]
                      for e in emag])
    if pos>=0:
        for i, x in enumerate(isa.pos_el_fe_induction):
            if x >= pos/180*np.pi:
                break
    else:
        demagmax = np.max(demag, axis=0)
        i = np.argmax(demagmax)
        x = isa.pos_el_fe_induction[i]

    hpol = demag[:, i]
    hmax = np.max(hpol)
    hpol[hpol == 0] = np.nan
    isa.rotate(isa.pos_el_fe_induction[i])
    _contour(ax, f'Demagnetization at pos. {round(x/np.pi*180, 1):.1f}°, '
             f'{isa.MAGN_TEMPERATURE} °C (max -{hmax:.1f} kA/m)',
             emag, hpol, '-H / kA/m', cmap=cmap, isa=isa,
             relminrange=relminrange)
    isa.rotate(0)
    logger.info("Max demagnetization %f kA/m", np.nanmax(hpol))


def __elements_of_subreg(isa, subreg):
    if subreg:
        if isinstance(subreg, list):
            sr = subreg
        else:
            sr = [subreg]
        for s in sr:
            yield from isa.get_subregion(s).elements()
    else:
        for e in isa.elements:
            yield e


def flux_density(isa, subreg=[], cmap=DEFAULT_CMAP,
                 relminrange=0.05, ax=0):
    """plot flux density of NC/I7/ISA7 model

    Args:
        isa: Isa7/NC object
        subreg: list of subregion names (all if empty)
        icur: cur index or last index if -1
        ibeta: beta angle index or last index if -1
        relminrange: rel minimum range for colorbar
    """
    elements = [e for e in __elements_of_subreg(isa, subreg)]
    fluxd = np.array([np.linalg.norm(e.flux_density()) for e in elements])
    _contour(ax, f'Flux Density T (max {np.max(fluxd):.1f} T)',
             elements, fluxd, '', cmap=cmap, relminrange=relminrange)
    logger.info("Max flux dens %f", np.max(fluxd))


def flux_density_eccentricity(isa, subreg=[], icur=-1, ibeta=-1,
                              cmap='plasma_r', ax=0, alpha=0.75,
                              relminrange=0.05):
    """plot eccentricity (axis ratio) of flux density in lamination

    Args:
        isa: Isa7/NC object
        subreg: list of subregion names (all if empty)
        icur: cur amplitude index or last index if -1
        ibeta: beta angle index or last index if -1
        relminrange: rel minimum range for colorbar
    """
    from ..utils import fft
    elements = []
    ecc = []
    pulsating = 0
    # eliminate double values at end
    # TODO: adapt to linear machines
    pos = [p
           for p in isa.pos_el_fe_induction
           if p < 2*np.pi/isa.pole_pairs] + [2*np.pi/isa.pole_pairs]
    i = len(pos)
    apos = np.array(pos)/np.pi*180
    for e in __elements_of_subreg(isa, subreg):
        if e.is_lamination():
            br = isa.el_fe_induction_1[e.key-1, :i+1, icur, ibeta]
            bt = isa.el_fe_induction_2[e.key-1, :i+1, icur, ibeta]
            brtmax = np.max(br-np.mean(br)), np.max(bt-np.mean(bt))
            if np.all(np.isclose(brtmax, 0)):
                continue
            elements.append(e)
            if np.any(np.isclose(brtmax, 0)):
                ecc.append(pulsating)
            else:
                br0 = fft(apos, br-np.mean(br))
                x = br0['a']*np.cos(2*np.pi*apos/br0['T0']+br0['alfa0'])
                bt0 = fft(apos, bt-np.mean(bt))
                y = bt0['a']*np.cos(2*np.pi*apos/bt0['T0']+bt0['alfa0'])
                if (br0['a'] > br0['nue'][isa.pole_pairs]
                    or bt0['a'] > bt0['nue'][isa.pole_pairs]):
                    ecc.append(pulsating)
                else:
                    kmax = np.argmax(np.linalg.norm((x, y), axis=0))
                    kmin = np.argmin(np.linalg.norm((x, y), axis=0))
                    a = np.linalg.norm((x[kmax], y[kmax]))
                    b = np.linalg.norm((x[kmin], y[kmin]))
                    ecc.append(b/a) #np.sqrt(1-b**2/a**2))

    _contour(ax, '', #'Eccentricity of Flux Density',
                 elements, ecc, 'axis ratio', cmap=cmap, alpha=alpha,
                 relminrange=relminrange)


def flux_density_pos(isa, ipos, subreg=[], icur=-1, ibeta=-1, cmap=DEFAULT_CMAP,
                     relminrange=0.05, ax=0):
    """plot flux density at rotor pos for each element of NC/I7/ISA7 model

    Args:
        isa: Isa7/NC object
        ipos: position index
        icur: cur index or last index if -1
        ibeta: beta angle index or last index if -1
        relminrange: rel minimum range for colorbar
    """
    elements = [e for e in __elements_of_subreg(isa, subreg)
                if e not in isa.airgap_center_elements]
    b = []
    for e in elements:
        fd = isa.flux_density(e, icur, ibeta)
        b.append(np.linalg.norm(
            (fd['bx'][ipos], fd['by'][ipos])))
    fluxd = np.array(b)
    pos = isa.pos_el_fe_induction[ipos]*180/np.pi
    isa.rotate(isa.pos_el_fe_induction[ipos])
    _contour(ax, f'Flux Density T at {pos:.1f}° (max {np.max(fluxd):.1f} T)',
             elements, fluxd, '', cmap=cmap, relminrange=relminrange)
    logger.info("Max flux dens %f", np.max(fluxd))
    isa.rotate(0)


def airgap_flux_density_pos(isa, ipos, icur=-1, ibeta=-1, ax=0):
    """plot flux density at rotor pos for each element of NC/I7/ISA7 model

    Args:
        isa: Isa7/NC object
        ipos: index of rotor position
        icur: index of current
        ibeta: index of beta
    """
    bx = []
    by = []
    pos = []
    for e in isa.airgap_center_elements:
        fd = isa.flux_density(e, icur, ibeta)
        bx.append(fd['bx'][ipos])
        by.append(fd['by'][ipos])
        pos.append(np.arctan2(e.center[1], e.center[0])/np.pi*100)
    if ax == 0:
        ax = plt.gca()
    ax.plot(pos, bx)
    ax.plot(pos, by)
    ax.grid()

    logger.info("Max flux dens %f", np.max(np.abs(bx)))


def loss_density(isa, subreg=[], cmap=DEFAULT_CMAP, relminrange=0.05, ax=0):
    """plot loss density of NC/I7/ISA7 model

    Args:
        isa: Isa7/NC object
        subreg: array of subregions (all if empty)
        cmap: color map
        relminrange: rel minimum range for colorbar
    """
    elements = [e for e in __elements_of_subreg(isa, subreg)]
    lossd = np.array([e.loss_density*1e-3 for e in elements])
    _contour(ax, 'Loss Density kW/m³', elements, lossd, '', cmap=cmap,
             relminrange=relminrange)


def temperature_distribution(isa, ax=0, cmap='plasma', relminrange=0.05):
    """plot temperature distribution of NC/I7/ISA7 model

    Args:
        isa: Isa7/NC object
        relminrange: rel minimum range for colorbar
    """
    temp = []
    elements = [e for e in isa.elements]
    for e in isa.elements:
        tmp = 0
        ctr = 1
        for n in e.vertices:
            tmp += n.vpot[-1]
            ctr = ctr + 1
        temp.append(tmp/ctr)
    _contour(ax, 'Temperature in K', elements, temp, '', cmap=cmap,
             relminrange=relminrange)

def punchdist(isa, cmap=DEFAULT_CMAP, ax=0):
    """plot punching border distances of NC/I7/ISA7 model
    Args:
      isa: Isa7/NC object
    """
    elam, pdist = isa.punchdist()
    _contour(ax, 'Punching Border Distances / mm)',
             elam, pdist*1e3, 'mm', cmap=cmap, isa=isa)



"""Creating plots

"""
from .machine import machine
from .fluxdens import airgap, airgap_fft
from .bch import torque, torque_fft, force, force_fft, \
    fluxdens_surface, winding_current, winding_flux, \
    voltage, voltage_fft, \
    pmrelsim, multcal, fasttorque, cogging, transientsc, \
    transientsc_demag, i1beta_torque, i1beta_ld, i1beta_lq, i1beta_psid, \
    i1beta_psiq, i1beta_psim, i1beta_up, \
    idq_ld, idq_lq, idq_psid, idq_psim, idq_psiq, idq_torque
from .char import mtpa, mtpv, characteristics, efficiency_map, losses_map
from .forcedens import forcedens, forcedens_surface, forcedens_contour, forcedens_fft
from .nc import spel, mesh, demag, demag_pos, \
    flux_density, flux_density_eccentricity, \
    airgap_flux_density_pos, loss_density
from .mcv import mcv_hbj, mcv_muer, felosses
from .phasor import i1beta_phasor, iqd_phasor, phasor
from .wdg import mmf, mmf_fft, zoneplan, winding_factors, winding, currdist
from .fieldlines import fieldlines



"""
Create longitudinal drawing of radial flux machine
"""
import matplotlib.pyplot as plt
import matplotlib.patches as pch

def _draw_shaft(ax, dy2, lfe):
    xx = (0.0, lfe, lfe, 0.0)
    yy = (dy2/2, dy2/2, -dy2/2, -dy2/2)
    ax.fill(xx, yy,
            facecolor='lightgrey',
            edgecolor='black',
            linewidth=0)
    xx = (-lfe/4, lfe+lfe/4, lfe+lfe/4, -lfe/4)
    yy = (dy2/4, dy2/4, -dy2/4, -dy2/4)
    ax.fill(xx, yy,
            facecolor='lightgrey',
            edgecolor='black',
            linewidth=0)

def _draw_rotor(ax, da1, dy2, lfe):
    ag = 0.02*da1
    xx = (0.0, lfe, lfe, 0.0)
    yy = (dy2/2, dy2/2, da1/2-ag, da1/2-ag)
    ax.fill(xx, yy,
            facecolor='skyblue',
            edgecolor='black',
            linewidth=0)
    yy = (-dy2/2, -dy2/2, -da1/2+ag, -da1/2+ag)
    ax.fill(xx, yy,
            facecolor='skyblue',
            edgecolor='black',
            linewidth=0)

def _draw_stator(ax, da1, dy1, lfe):
    xx = (0.0, lfe, lfe, 0.0)
    yy = (da1/2, da1/2, dy1/2, dy1/2)
    # yoke
    ax.fill(xx, yy,
            facecolor='skyblue',
            edgecolor='black',
            linewidth=0)

    yy = (-da1/2, -da1/2, -dy1/2, -dy1/2)
    ax.fill(xx, yy,
            facecolor='skyblue',
            edgecolor='black',
            linewidth=0)

    # winding
    yh = (dy1-da1)/2
    xx = (-yh/2, 0, 0, -yh/2)
    yy = (da1/2, da1/2, dy1/2-yh/2, dy1/2-yh/2)
    ax.fill(xx, yy, facecolor='gold',
                edgecolor='black',
                linewidth=0)

    xx = (lfe, lfe+yh/2, lfe+yh/2, lfe)
    ax.fill(xx, yy, facecolor='gold',
                edgecolor='black',
                linewidth=0)

    yy = (-da1/2, -da1/2, -dy1/2+yh/2, -dy1/2+yh/2)
    ax.fill(xx, yy, facecolor='gold',
                edgecolor='black',
                linewidth=0)

    xx = (-yh/2, 0, 0, -yh/2)
    ax.fill(xx, yy, facecolor='gold',
                edgecolor='black',
                linewidth=0)

def machine(machine, ax):
    dy2 = machine['inner_diam']*1e3
    dy1 = machine['outer_diam']*1e3
    da1 = machine['bore_diam']*1e3
    lfe = machine['lfe']*1e3

    _draw_rotor(ax, da1, dy2, lfe)
    _draw_stator(ax, da1, dy1, lfe)
    _draw_shaft(ax, dy2, lfe)
    ax.set_aspect('equal')

    for loc, spine in ax.spines.items():
        spine.set_color('none')  # don't draw spine
    #ax.yaxis.set_ticks([])
    #ax.xaxis.set_ticks([])


if __name__ == '__main__':
    machine1 = {
        "outer_diam": 0.2442,
        "bore_diam": 0.179,
        "inner_diam": 0.06,
        "airgap": 0.7e-3,
        "lfe": 0.083,
    }
    fig, ax = plt.subplots()
    machine(machine1, ax)
    plt.show()



"""
    femagtools.plot.forcedens
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Creating force density plots


"""
import numpy as np
import matplotlib.pyplot as plt
try:
    from matplotlib import colormaps
    cmap_viridis = colormaps['viridis']
    cmap_jet = colormaps['jet']
    cmap_YlOrBr = colormaps['YlOrBr']
except:  # older matplotlib
    from matplotlib import cm
    cmap_viridis = cm.viridis
    cmap_jet = cm.jet
    cmap_YlOrBr = cm.YlOrBr


def _create_3d_axis():
    """creates a subplot with 3d projection if one does not already exist"""
    from matplotlib.projections import get_projection_class
    from matplotlib import _pylab_helpers

    create_axis = True
    if _pylab_helpers.Gcf.get_active() is not None:
        if isinstance(plt.gca(), get_projection_class('3d')):
            create_axis = False
    if create_axis:
        plt.figure()
        plt.subplot(111, projection='3d')


def _plot_surface(ax, x, y, z, labels, azim=None, cmap=cmap_viridis):
    """helper function for surface plots"""
    # ax.tick_params(axis='both', which='major', pad=-3)
    assert np.size(x) > 1 and np.size(y) > 1 and np.size(z) > 1
    if azim is not None:
        ax.azim = azim
    X, Y = np.meshgrid(x, y)
    Z = np.ma.masked_invalid(z)
    ax.plot_surface(X, Y, Z,
                    rstride=1, cstride=1,
                    cmap=cmap,
                    vmin=np.nanmin(z), vmax=np.nanmax(z),
                    linewidth=0, antialiased=True)
#                    edgecolor=(0, 0, 0, 0))
    # ax.set_xticks(xticks)
    # ax.set_yticks(yticks)
    # ax.set_zticks(zticks)
    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])
    ax.set_title(labels[2])


def forcedens(title, pos, fdens, ax=0):
    """plot force densities"""
    if ax == 0:
        ax = plt.gca()
    ax.set_title(title)
    ax.grid(True)

    ax.plot(pos, [1e-3*ft for ft in fdens[0]], label='F tang')
    ax.plot(pos, [1e-3*fn for fn in fdens[1]], label='F norm')
    ax.legend()
    ax.set_xlabel('Pos / deg')
    ax.set_ylabel('Force Density / kN/m²')


def forcedens_surface(fdens, attr='FN', ax=0, cmap=cmap_jet):
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    z = 1e-3*np.array(getattr(fdens, attr))
    n, m = z.shape
    x = np.linspace(0, 360, m)
    y = np.linspace(0, 360, n)
    _plot_surface(ax, x, y, z,
                  ('Rotor pos/°', 'Pos/°', f'{attr} / kN/m²'),
                  cmap=cmap)

def forcedens_contour(fdens, attr='FN', ax=0, cmap=cmap_jet):
    if ax == 0:
        ax = plt.gca()
    z = 1e-3*np.array(getattr(fdens, attr))
    n, m = z.shape
    x = np.linspace(0, 360, m)
    y = np.linspace(0, 360, n)
    cs = ax.contourf(x, y, z, cmap=cmap)
    ax.set_xlabel('Rotor pos/°')
    ax.set_ylabel('Pos/°')
    if attr == 'FN':
        ax.set_title('Radial Force Density')
    elif attr == 'FT':
        ax.set_title('Tangential Force Density')
    cfig = ax.get_figure()
    cbar = cfig.colorbar(cs, ax=ax)
    cbar.ax.set_ylabel('kN/m²')

def forcedens_fft(title, fdens, harmmax=(), #(200, 40),
                  cmap=cmap_YlOrBr,
                  ax=0):
    """plot force densities FFT
    Args:
      title: plot title
      fdens: force density object
      cmap: colormap
    """
    if ax == 0:
        ax = plt.axes(projection="3d")

    FN = 1e-3*fdens.fft()['fn_harm']['amplitude']
    if harmmax:
        shape = np.min((harmmax, np.shape(FN)), axis=0)
    else:
        shape = np.shape(FN)

    _x = np.arange(shape[0])+1
    _y = np.arange(shape[1])+1
    _xx, _yy = np.meshgrid(_x, _y)
    x, y = _xx.ravel(), _yy.ravel()
    top = np.ravel(FN[:shape[0],:shape[1]])
    bottom = np.zeros_like(top)
    maxy = max(x[-1], y[-1])
    width, depth = 5*x[-1]/maxy, 5*y[-1]/maxy

    min_height = np.min(top)
    max_height = np.max(top)
    rgba = [cmap((k-min_height)/max_height) for k in top]
    ax.bar3d(x, y, bottom, width, depth, top, color=rgba)

    ax.set_xlim(0, shape[0]+1)
    ax.set_ylim(0, shape[1]+1)
    ax.set_title(title)
    ax.set_xlabel('M')
    ax.set_ylabel('N')
    ax.set_zlabel('kN/m²')



"""running FEMAG

"""
import subprocess
import pathlib
import os
import glob
import logging
try:
    import zmq
except ImportError:
    pass
import sys
import json
import io
import time
import re
import threading
import femagtools.model
import femagtools.magnet
import femagtools.conductor
import femagtools.windings
import femagtools.mcv
import femagtools.asm
import femagtools.airgap as ag
import femagtools.fsl
import femagtools.config
import femagtools.ecloss
import femagtools.forcedens
import femagtools.zmq

from femagtools import ntib


logger = logging.getLogger(__name__)


class FemagError(Exception):
    pass


def handle_process_output(filedes, outfile, log):
    """read from file descriptor and direct lines to logger and outfile"""
    with open(outfile, 'wb') as fp:
        for line in filedes:
            fp.write(line)
            if log:
                if (b'' == line or
                    b'\x1b' in line or  # ignore terminal escape seq
                    b'\n' == line or
                    b'Start:' in line or
                    b'Stop:' in line or
                        b'Elapsed time:' in line):
                    continue
                if line:
                    try:
                        logger.info(" > %s",
                                    line.decode().strip())
                    except UnicodeDecodeError:
                        # ignore
                        pass


def set_magnet_properties(model, simulation, magnets):
    """set temperature adapted magnet properties"""
    if not hasattr(model, 'magnet'):
        return
    if isinstance(simulation, dict):
        if 'hc_min' in simulation:  # backward compatibility
            model.hc_min = simulation['hc_min']
        if 'magn_temp' not in simulation:
            return
        magn_temp = simulation['magn_temp']
    else:
        if hasattr(simulation, 'hc_min'):  # backward compatibility
            model.hc_min = simulation.hc_min
        if not hasattr(simulation, 'magn_temp'):
            return
        magn_temp = simulation.magn_temp
    if 'material' not in model.magnet:
        return

    try:
        material = model.magnet['material']
        magnetMat = magnets.find(material)
        if magnetMat:
            model.magnet['temp_prop'] = {
                'remanenc': magnetMat.get('remanenc', 1.2),
                'magntemp': magn_temp
            }
            relperm = magnetMat.get('relperm', 1.05)
            tempcoefmuer = 1
            tempcoefbr = magnetMat.get('temcoefbr', 0)
            tempcoefhc = magnetMat.get('temcoefhc', 0)
            if tempcoefbr and tempcoefhc:
                tempcoefmuer = (tempcoefbr-tempcoefhc)/(1+tempcoefhc*(magn_temp-20))
            if 'temcoefmuer' in magnetMat:
                tempcoefmuer = magnetMat['temcoefmuer']
            if not hasattr(model, 'hc_min'):
                for k in 'Hk', 'HcJ':
                    if k in magnetMat:
                        hcmin = magnetMat.get(k)
                        model.hc_min = hcmin * (1+tempcoefhc*(magn_temp-20))
                        break
            if hasattr(model, 'hc_min'):
                logger.info("hc_min %f", model.hc_min)
            model.magnet['temp_prop']['relperm'] = \
                (1+tempcoefmuer*(magn_temp-20))*relperm
            if tempcoefbr:
                model.magnet['temp_prop']['temcoefbr'] = tempcoefbr
            if tempcoefhc:
                model.magnet['temp_prop']['temcoefhc'] = tempcoefhc
    except AttributeError:
        pass


class BaseFemag(object):
    def __init__(self, workdir, cmd, magnetizingCurves, magnets, condMat,
                 templatedirs=[]):
        self.workdir = workdir
        self.magnets = []
        self.magnetizingCurves = []
        self.condMat = []
        self.templatedirs = templatedirs
        if cmd:
            self.cmd = cmd
        else:
            self.cmd = ''

        if magnetizingCurves:
            if isinstance(magnetizingCurves,
                          femagtools.mcv.MagnetizingCurve):
                self.magnetizingCurves = magnetizingCurves
            else:
                self.magnetizingCurves = femagtools.mcv.MagnetizingCurve(
                    magnetizingCurves)

        if magnets:
            if isinstance(magnets, femagtools.magnet.Magnet):
                self.magnets = magnets
            else:
                self.magnets = femagtools.magnet.Magnet(magnets)

        if condMat:
            if isinstance(condMat, femagtools.conductor.Conductor):
                self.condMat = condMat
            else:
                self.condMat = femagtools.conductor.Conductor(condMat)

    def copy_winding_file(self, name, wdg):
        wdg.write(name, self.workdir)

    def copy_magnetizing_curves(self, model, dir=None, recsin='', feloss=''):
        """extract mc names from model and write files into workdir or dir if given

        Return:
            list of extracted mc names (:obj:`list`)
        """
        dest = dir if dir else self.workdir
        if isinstance(feloss, (int, float)):
            try:
                feloss = {1: 'jordan', 2: "steinmetz", 10:"modified steinmetz", 11: 'bertotti'}[int(feloss)]
            except KeyError:
                feloss = ''
        return [self.magnetizingCurves.writefile(m[0], dest,
                                                 fillfac=m[1],
                                                 recsin=recsin,
                                                 feloss=feloss)
                for m in model.set_magcurves(
            self.magnetizingCurves, self.magnets)]

    def create_wdg_def(self, model):
        name = 'winding'
        w = femagtools.windings.Winding(
            dict(
                Q=model.stator['num_slots'],
                p=model.poles//2,
                m=len(model.winding['wdgdef']),
                windings=model.winding['wdgdef']))
        self.copy_winding_file(name, w)
        return name

    def create_fsl(self, machine, simulation):
        """create list of fsl commands
        Args:
          machine: dict of fe machine model (stator, rotor/magnet, winding)
          simulation: dict of FE simulation (can be empty)
        """
        self.model = femagtools.model.MachineModel(machine)
        self.modelname = self.model.name
        recsin = ''
        feloss = ''
        if simulation:
            recsin = simulation.get('recsin', '')
            feloss = simulation.get('calc_fe_loss', '')
        self.copy_magnetizing_curves(self.model, recsin=recsin, feloss=feloss)

        if feloss and \
            int(feloss) in (1, 2):
            simulation["calc_fe_loss"] = 1

        try:
            if 'wdgdef' in self.model.winding:
                self.model.winding['wdgfile'] = self.create_wdg_def(
                    self.model)
        except AttributeError:
            pass
        builder = femagtools.fsl.Builder(self.templatedirs)
        if simulation:
            if 'num_par_wdgs' not in simulation:
                try:
                    num_par_wdgs = self.model.winding['num_par_wdgs']
                    simulation['num_par_wdgs'] = num_par_wdgs
                except:
                    pass
            set_magnet_properties(self.model, simulation, self.magnets)
            return builder.create(self.model, simulation,
                                  self.magnets, self.condMat)
        return builder.create_model(self.model,
                                    self.magnets,
                                    self.condMat) + ['save_model("cont")']

    def get_log_value(self, pattern, modelname='FEMAG-FSL.log'):
        result = []
        pat = re.compile(pattern)
        with open(os.path.join(self.workdir, 'FEMAG-FSL.log')) as f:
            for line in f:
                m = pat.search(line)
                if m:
                    result.append(float(line.split(':')[-1].split()[0]))
        return result

    def get_bch_file(self, modelname, offset=0):
        return self.get_result_file(modelname, 'B(AT)?CH', offset)

    def get_plt_file(self, modelname, offset=0):
        return self.get_result_file(modelname, 'PLT0', offset)

    def get_asm_file(self, modelname, offset=0):
        return self.get_result_file(modelname, 'ASM', offset)

    def get_ts_files(self, modelname, offset=0):
        return self.get_result_file(modelname, 'TS', offset)

    def get_result_file(self, modelname, ext, offset=0):
        """return latest result (bch, asm) file (if any)"""
        filelist = glob.glob(os.path.join(
            self.workdir, modelname+'_[0-9][0-9][0-9].*'))
        if (filelist):
            return sorted([f for f in filelist
                           if re.search(ext, f.split('.')[-1])])[-1-offset]
        return ''

    def get_result_file_list(self, modelname, ext):
        """return list of result (me) file (if any)"""
        filelist = sorted(glob.glob(os.path.join(
            self.workdir, modelname+'_[0-9][0-9][0-9].'+ext)))
        if len(filelist) > 0:
            return filelist
        return []

    def read_asm(self, modelname=None, offset=0):
        "read most recent ASM file and return result"
        if not modelname:
            modelname = self._get_modelname_from_log()

        asmfile = self.get_asm_file(modelname, offset)
        if asmfile:
            logger.info("Read ASM {}".format(asmfile))
            return femagtools.asm.read(asmfile)
        return {}

    def read_ts(self, modelname=None):
        "read most recent TS files and return result"
        import femagtools.ts
        if not modelname:
            modelname = self._get_modelname_from_log()

        logger.info("Read TS {}".format(modelname))
        return femagtools.ts.read_st(self.workdir, modelname)

    def read_modal(self, modelname=None):
        "read modal analysis output and return result"
        import femagtools.me
        filelist = self.get_result_file_list(modelname + '_Mode', 'txt')
        psfiles = self.get_result_file_list(modelname + '_Mode', 'ps')
        if len(filelist) > 0:
            logger.info("Read Eigenvectors {}".format(modelname))
            return femagtools.me.get_eigenvectors(filelist, psfiles)
        else:
            return ''

    def read_bch(self, modelname=None, offset=0):
        "read most recent BCH/BATCH file and return result"
        # read latest bch file if any
        if not modelname:
            modelname = self._get_modelname_from_log()

        result = femagtools.bch.Reader()
        bchfile = self.get_bch_file(modelname, offset)
        if bchfile:
            logger.info("Read BCH {}".format(bchfile))
            with io.open(bchfile, encoding='latin1',
                         errors='ignore') as f:
                result.read(f)
        return result

    def read_forcedens(self, modelname=None, offset=0):
        "read most recent PLT0 file and return result"
        if not modelname:
            modelname = self._get_modelname_from_log()

        pltfile = self.get_plt_file(modelname, offset)
        if pltfile:
            logger.info("Read PLT0 {}".format(pltfile))
            return femagtools.forcedens.read(pltfile)
        return None

    def read_isa(self, modelname=None):
        "read most recent I7/ISA7 file and return result"
        if not modelname:
            modelname = self._get_modelname_from_log()
        import femagtools.isa7
        return femagtools.isa7.read(os.path.join(self.workdir, modelname))

    def read_nc(self, modelname=None):
        "read most recent NC file and return result"
        if not modelname:
            modelname = self._get_modelname_from_log()
        import femagtools.nc
        return femagtools.nc.read(os.path.join(self.workdir, modelname))

    def read_los(self, modelname=None):
        "read most recent LOS file and return result"
        # read latest los file if any
        if not modelname:
            modelname = self._get_modelname_from_log()

        losfile_list = sorted(glob.glob(os.path.join(
            self.workdir, modelname+'_[0-9][0-9][0-9].LOS')))
        if len(losfile_list) > 0:
            return ntib.read_los(losfile_list[-1])

        return dict()

    def read_hsn(self, modelname=None):
        import numpy as np
        "read heat network result"
        _map = {
            "StZa": "plfe1",
            "outs": "-",
            "StJo": "plfe1",
            "Slot": "-",
            "Shaf": "-",
            "Iron": "plfe2",
            "PMag": "plmag",
            "PMag_1": "plmag",
            "PMag_2": "plmag",
            "PMag_3": "plmag",
            "PMag_4": "plmag",
            "W1  ":   "plcu1",
            "W2  ":   "plcu1",
            "W3  ":   "plcu1"
        }
        if not modelname:
            modelname = self._get_modelname_from_log()
        hsn_list = sorted(glob.glob(os.path.join(
            self.workdir, modelname+'.hsn')))
        with open(hsn_list[-1], 'r') as f:
            hsn_data = json.load(f)

        # area calculation
        nc_file = self.read_nc(modelname)
        slot_area = 0.0
        wdg_area = []
        for i in nc_file.windings:
            for j in i.subregions:
                wdg_area.append(j.area())

        slot_area = np.sum(wdg_area).item()/3
        magnet_area = 0.0
        num_sreg_mag = 0
        area = dict()
        for i in nc_file.subregions:
            if i.name in ("StZa", "StJo", "Iron"):
                area[i.name] = i.area().item()
            elif i.name == "PMag":
                magnet_area += i.area().item()
                num_sreg_mag += 1
            else:
                pass

        area['PMag'] = magnet_area/num_sreg_mag
        for i in ('W1', 'W2', 'W3'):
            area[i] = slot_area

        pmag_index = []
        if "Nodes" in hsn_data:
            for k ,i in enumerate(hsn_data['Nodes']):
                i.update({"mass": i['weight'], "losses": _map[i['Name']]})
                if "PMag" in i['Name']:
                    pmag_index.append(k)
                if i['Name'].strip() in area.keys():
                    i.update({"area": area[i['Name'].strip()]})
            if pmag_index:
                for i in range(len(pmag_index)):
                    hsn_data["Nodes"][pmag_index[i]]['Name'] = f"PMag_{i+1}"
        with open(hsn_list[-1], 'w') as f:
            json.dump(hsn_data, f)
        return nc_file

    def _get_modelname_from_log(self):
        """
        Read the modelname from the Femag Log file
        """
        try:
            with open(os.path.join(self.workdir, 'FEMAG-FSL.log')) as f:
                for l in f:
                    if l.startswith('New model') or l.startswith('Load model'):
                        return l.split('"')[1]
        except FileNotFoundError:
            pass

        return sorted(pathlib.Path(self.workdir).glob('*.PROT'),
                      key=lambda x: x.stat().st_mtime, reverse=True)[0].stem

    def readResult(self, machine, simulation, bch=None):
        if simulation:
            if simulation['calculationMode'] == "fieldcalc":
                nc = self.read_nc()
                pmod = nc.poles_sim
                r = {'airgap': ag.read(
                    os.path.join(self.workdir, 'bag.dat'), pmod=pmod)}
                if 'plots' in simulation:
                    if 'field_lines' in simulation['plots']:
                        r['field_lines'] = os.path.join(
                            self.workdir, 'field.svg')
                return r
            if simulation['calculationMode'] == "pm_sym_loss":
                return self.read_los(self.modelname)

            if simulation['calculationMode'] == 'asyn_motor':
                return self.read_asm(self.modelname)

            if simulation['calculationMode'] == 'calc_field_ts':
                return self.read_ts(self.modelname)

            if simulation['calculationMode'] == 'modal_analysis':
                return self.read_modal(self.modelname)

            if simulation['calculationMode'] == 'therm-dynamic':
                temp = [[float(n) for n in l.split()]
                        for l in (pathlib.Path(self.workdir) /
                                  'temperature.dat').read_text().split('\n') if l]
                ttemp = list(zip(*temp))
                return {'t': ttemp[0], 'temperature': ttemp[1]}

            if simulation['calculationMode'] == 'hsn':
                model = None
                try:
                    model = self.read_hsn()
                except:
                    pass
                if model is None:
                    model = self.read_nc()
                return model.get_minmax_temp()

            if not bch:
                bch = self.read_bch(self.modelname)

            if 'airgap_induc' in simulation:
                try:
                    pmod = bch.machine['p_sim']
                except KeyError:
                    pmod = 0
                bch.airgap = ag.read(os.path.join(self.workdir, 'bag.dat'),
                                     pmod=pmod)

            if simulation['calculationMode'] == 'pm_sym_fast' or \
                simulation['calculationMode'] == 'torq_calc':
                if simulation.get('shortCircuit', False):
                    from .shortcircuit import shortcircuit
                    set_magnet_properties(self.model, simulation, self.magnets)
                    bch.scData = shortcircuit(self,
                                              {'name': self.model.name,
                                               'poles': self.model.poles,
                                               'hc_min': self.model.get('hc_min', 0)},
                                               bch, simulation)
                    #bch.torque += bchsc.torque
                    #bch.demag += bchsc.demag

            if simulation.get('magnet_loss', False):
                logger.info('Evaluating magnet losses...')
                ops = range(len(bch.torque))
                m = femagtools.ecloss.MagnLoss(
                    nc=self.read_nc(), ibeta=ops)
                try:
                    # change from ialh to ialh2: since v1.8.1
                    magn_losses = m.calc_losses_ialh2()
                except:
                    magn_losses = list(range(len(ops)))

                if len(ops) != len(bch.losses):
                    magn_losses.insert(0, magn_losses[0])

                magn_losses = [float(i) for i in magn_losses]
                try:
                    for i in range(len(bch.losses)):
                        bch.losses[i].update({"magnetH": magn_losses[i]})
                except:
                    pass
                # pass losses to bch object for th usage
                try:
                    bch.magnet_loss_th = m.th_loss
                except:
                    pass
            try:
                if hasattr(self, 'dy2'):
                    setattr(bch, 'dy2', self.dy2)
            except:
                pass
            return bch


class Femag(BaseFemag):
    """Invoke and control execution of FEMAG

    Args:
        workdir: name of working directory
        cmd: name of femag program (default is config.executable)
        magnetizingCurves: collection of lamination material curves or name of directory
        magnets: collection of magnet material
        condMat: collection of conductor material
        templatedirs: (list) names of directories that include mako files as fsl templates

    """

    def __init__(self, workdir, cmd=None, templatedirs=[],
                 magnetizingCurves='.', magnets=None, condMat=[]):
        super(self.__class__, self).__init__(workdir, cmd,
                                             magnetizingCurves, magnets, condMat,
                                             templatedirs=templatedirs)

    def run(self, filename, options=['-b'], fsl_args=[],
            stateofproblem='mag_static'):
        """invoke FEMAG in current workdir

        Args:
            filename: name of FSL file to execute
            options: list of FEMAG options
            fsl_args: list of FSL argument options
            stateofproblem: (str) one of config.state_of_problem_set
        Raises:
            FemagError
        """
        if self.cmd:
            cmd = self.cmd
        else:
            cmd = femagtools.config.get_executable(stateofproblem)
        if (cmd.find('wfemag') > -1 and
            '-b' in options and
                '-m' not in options):
            options.insert(0, '-m')
        args = [cmd] + options + [filename] + fsl_args

        basename = pathlib.Path(filename).name
        outname = os.path.join(self.workdir, basename+'.out')
        errname = os.path.join(self.workdir, basename+'.err')
        with open(errname, 'w') as err:
            logger.info('invoking %s', ' '.join([str(a) for a in args]))
            proc = subprocess.Popen(
                args,
                stdout=subprocess.PIPE, stderr=err, cwd=self.workdir)
            stdoutthr = threading.Thread(target=handle_process_output,
                                         args=(proc.stdout, outname, True))
            stdoutthr.start()
        proc.wait()
        stdoutthr.join()
        errs = []
        # print femag output
        with io.open(outname, encoding='latin1', errors='ignore') as outfile:
            errLine = False
            for l in outfile:
                if l.find('ERROR') > -1:
                    errs.append(l.strip())
                    errLine = True
                elif errLine and l.startswith(' '):  # additional error line
                    errs.append(l)
                else:
                    errLine = False

        rc = proc.returncode
        logger.info("%s exited with returncode %d (num errs=%d)",
                    cmd, rc, len(errs))
        if rc != 0 or errs:
            with io.open(errname, encoding='latin1',
                         errors='ignore') as errfile:
                for l in errfile:
                    errs.append(l.strip())
            errs.insert(0, 'Exit code {}'.format(rc))
            raise FemagError("\n".join(errs))

    def cleanup(self):
        "removes all created files in workdir"
        if not os.path.exists(self.workdir):
            return
        cleanfiles = ('*.B*CH', '*.I*7-*', '*.A*7-*', '*.nc-*',
                      '*.dat', '*.LOS', '*.svg', '*.png', '*.hxy')
        # '*.TMC','*.TMO', '*.PROT'):
        for p in cleanfiles:
            for f in glob.glob(os.path.join(self.workdir, p)):
                os.remove(f)

    def __call__(self, machine, simulation={},
                 options=['-b'], fsl_args=[], fslfile='femag.fsl'):
        """setup fsl file, run calculation and return
        BCH, ASM, TS or LOS results if any."""
        with open(os.path.join(self.workdir, fslfile), 'w') as f:
            f.write('\n'.join(self.create_fsl(machine,
                                              simulation)))
        if simulation:
            stateofproblem = simulation.get('stateofproblem', 'mag_static')
            if 'eccentricity' in simulation:
                #assert (self.model.stator['num_slots']
                #        == self.model.stator['num_slots_gen'])
                for k in ('bore_diam', 'airgap'):
                    if k not in simulation['eccentricity']:
                        simulation['eccentricity'][k] = self.model.get(k)

            if 'poc' in simulation:
                with open(os.path.join(self.workdir,
                                       simulation['pocfilename']), 'w') as f:
                    f.write('\n'.join(simulation['poc'].content()))
            if simulation['calculationMode'] == "pm_sym_loss":
                with open(os.path.join(self.workdir,
                                       self.modelname+'.ntib'), 'w') as f:
                    f.write('\n'.join(ntib.create(
                        simulation['speed'],
                        simulation['current'],
                        simulation['angl_i_up'])))
                # TODO: add r1, m
            if simulation['calculationMode'] == 'therm-dynamic':
                load = simulation['load']
                (pathlib.Path(self.workdir)/'load.csv').write_text(
                    '\n'.join([','.join([f"{load[k][i]}"
                                         for k in ('t', 'n', 'T', 'i1', 'beta')])
                               for i in range(0, len(load['t']))]))
        else:
            stateofproblem = 'mag_static'

        self.run(fslfile, options, fsl_args, stateofproblem=stateofproblem)

        try:
            setattr(self, "dy2", machine['stator']['dy2'])
        except:
            pass

        if simulation:
            return self.readResult(machine, simulation)

        return {'status': 'ok', 'message': self.modelname,
                'model': self.model.props()}

class FemagTask(threading.Thread):
    def __init__(self, port, args, workdir, logdir):
        threading.Thread.__init__(self)
        self.port = port
        self.args = args + [str(self.port)]
        self.proc = None
        self.returncode = None
        self.workdir = workdir
        self.logdir = logdir

    def run(self):
        logger.info("femag is ready on port %d workdir %s",
                    self.port, self.workdir)
        outname = os.path.join(self.logdir, f'femag-{self.port}.out')
        errname = os.path.join(self.logdir, f'femag-{self.port}.err')
        with open(outname, 'w') as out, open(errname, 'w') as err:
            self.proc = subprocess.Popen(
                self.args,
                stdout=out, stderr=err, cwd=self.workdir)

        self.returncode = self.proc.wait()


class ZmqFemag(BaseFemag):
    """Invoke and control execution of FEMAG with ZeroMQ

    Args:
        port: port number of req socket
        host: hostname (ip addr)
        workdir: name of working directory
        logdir: name of logging directory (default is workdir/log)
        cmd: name of femag program
        templatedirs: (list) names of directories that include mako files as fsl templates
    """

    def __init__(self, port, host='localhost', workdir='', logdir='',
                 cmd=None,
                 magnetizingCurves=None, magnets=None, condMat=[],
                 templatedirs=[]):
        super(self.__class__, self).__init__(
            workdir, cmd,
            magnetizingCurves, magnets, condMat,
            templatedirs=templatedirs)
        self.host = host
        self.port = port
        self.femaghost = ''
        self.logdir = logdir if logdir else os.path.join(workdir, 'log')
        if not os.path.exists(self.logdir):
            os.makedirs(self.logdir)
        self.request_socket = None
        self.subscriber = None

    def close(self):
        if self.subscriber:
            self.subscriber.stop()

        if self.request_socket:
            logger.debug("close request_socket")
            self.request_socket.close()
            self.request_socket = None

        logger.debug("done")

    def __del__(self):
        try:
            self.close()
        except AttributeError:
            pass
        logger.debug("Destructor ZmqFemag")

    def stopFemag(self):
        if self.femagTask and self.femagTask.proc:
            logger.info("stopFemagTask")
            self.femagTask.proc.kill()
            self.femagTask.proc.wait()
            self.femagTask = None
            self.request_socket = None
            self.subscriber.stop()
            self.subscriber = None
            logger.info("stopFemagTask Done")
        else:
            logger.warning("stopFemag not implemented")

    def __req_socket(self):
        """returns a new request client"""
        context = zmq.Context.instance()
        request_socket = context.socket(zmq.REQ)
        url = 'tcp://{0}:{1}'.format(
            self.host, self.port)
        logger.debug("connect req socket %s", url)
        request_socket.connect(url)
        return request_socket

    def subscribe(self, notify):
        """attaches a notify function"""
        logger.info("Subscribe on '%s' port %d", self.femaghost, self.port+1)
        femagtools.zmq.SubscriberTask.clear()
        if self.subscriber is None:
            # progress/xyplot at a configured timestep published
            header = [b'progress', b'xyplot', b'license']
            self.subscriber = femagtools.zmq.SubscriberTask(
                port=self.port+1, host=self.femaghost, notify=notify, header=header)
            self.subscriber.start()
        else:
            # reattach?
            self.subscriber.notify = notify
        return

    def __is_running(self):
        try:
            return self.femagTask.proc and self.femagTask.returncode is None
        except:
            pass
        return False

    def send_request(self, msg, pub_consumer=None, timeout=None):
        if not self.request_socket:
            self.request_socket = self.__req_socket()
        if not msg:
            return [b'{"status":"ignored",{}}']
        if timeout:
            self.request_socket.setsockopt(zmq.RCVTIMEO, timeout)
            self.request_socket.setsockopt(zmq.LINGER, 0)
        else:
            self.request_socket.setsockopt(zmq.RCVTIMEO, -1)  # default value

        for m in msg[:-1]:
            self.request_socket.send_string(m, flags=zmq.SNDMORE)
        if isinstance(msg[-1], list):
            self.request_socket.send_string('\n'.join(msg[-1]))
        else:
            self.request_socket.send_string(msg[-1])

        errmsg = ''
        max_num_again = 2
        again = 0
        while True:
            try:
                return self.request_socket.recv_multipart()
            except zmq.error.Again as e:
                again += 1
                # logger.exception("send_request")
                errmsg = str(e)
                logger.warning("send_request: %s Message %s, host: %s port: %s", str(
                    e), msg, self.host, self.port)
                if again >= max_num_again:
                    break
                continue
        logger.info("send_request failed %s", errmsg.encode())
        return [b'{"status":"error", "message":"' + errmsg.encode() + b'"}']

    def send_fsl(self, fsl, timeout=None):
        """sends FSL commands in ZMQ mode and blocks until commands are processed

        Args:
            fsl: string (or list) of FSL commands
            timeout: The maximum time (in milliseconds) to wait for a response

        Return:
            status
        """
        header = 'FSL'
        self.request_socket.send_string(header, flags=zmq.SNDMORE)

        if isinstance(fsl, list):
            self.request_socket.send_string('\n'.join(fsl))
        else:
            self.request_socket.send_string(fsl)
        logger.debug("Sent fsl wait for response")

        if timeout:
            self.request_socket.setsockopt(zmq.RCVTIMEO, timeout)
            self.request_socket.setsockopt(zmq.LINGER, 0)
        else:
            self.request_socket.setsockopt(zmq.RCVTIMEO, -1)  # default value
        import datetime
        startTime = datetime.datetime.now() if timeout else None
        while True:
            try:
                response = self.request_socket.recv_multipart()
                # NOTE: femag encoding is old school
                return [s.decode('latin1') for s in response]
            except zmq.error.Again as e:
                logger.info("Again [%s], timeout: %d", str(e), timeout)
                if not startTime:
                    continue

                diffTime = datetime.datetime.now() - startTime
                logger.debug("Diff msec[%d]", diffTime.microseconds)
                if diffTime.microseconds < timeout:
                    continue

                logger.info("ALERT not running close socket")
                self.close()
                return ['{"status":"error", "message":"Femag is not running"}', '{}']

            except Exception as e:
                logger.exception("send_fsl")
                logger.error("send_fsl: %s", str(e))
                if timeout:  # only first call raises zmq.error.Again
                    return ['{"status":"error", "message":"Femag is not running"}', '{}']
                msg = json.dumps(str(e))
                return ['{"status":"error", "message":'+msg+'}', '{}']

    def run(self, options=['-b'], restart=False, procId=None,
            stateofproblem='mag_static'):  # noqa: C901
        """invokes FEMAG in current workdir and returns pid

        Args:
            options: list of FEMAG options
            stateofproblem: str one of config.executable
        """
        if self.__is_running():
            if restart:
                logger.info("must restart")
                self.quit(True)
                self.femagTask.join()
                logger.info("Stopped procId: %s", self.femagTask.proc.pid)
            else:
                return self.femagTask.proc.pid

        if self.cmd:
            cmd = self.cmd
        else:
            cmd = femagtools.config.get_executable(stateofproblem)
        if (cmd.find('wfemag') > -1 and
            '-b' in options and
                '-m' not in options):
            options.insert(0, '-m')

        args = [cmd] + options
        self.femagTask = FemagTask(self.port, args, self.workdir, self.logdir)
        self.femagTask.start()
        if not self.request_socket:
            self.request_socket = self.__req_socket()

        # check if mq is ready for listening
        lcount = 300
        for t in range(lcount):
            time.sleep(0.1)
            if self.__is_running():
                if self.femagTask.proc:
                    logger.info("femag (pid: '{}') is listening".format(
                        self.femagTask.proc.pid))
                break

        return self.femagTask.proc.pid if self.femagTask.proc else 0

    def quit(self, save_model=False):
        """terminates femag"""

        if not self.__is_running():
            logger.info("Femag already stopped")
            return

        if self.subscriber:
            self.subscriber.stop()
            self.subscriber = None

        # send exit flags
        f = '\n'.join(['exit_on_end = true',
                       'exit_on_error = true'])
        response = self.send_fsl(f)

        # send quit command
        try:
            response = [r.decode('latin1')
                        for r in self.send_request(
                ['CONTROL', 'quit'], timeout=2000)]
        except Exception as e:
            logger.error("Femag Quit zmq message %s", e)

        logger.debug("Sent QUIT to femag %s", response)
        # if query, send a answer
        obj = json.loads(response[0])
        logger.debug("status: {}".format(obj['status']))
        if obj['status'] == 'Query':
            logger.info('query: %s => %s',
                        obj['message'], 'saved' if save_model else 'not saved')

            # Only send one msg
            response = self.request_socket.send_string(
                'Ok' if save_model else 'Cancel')

    def upload(self, files):
        """upload file or files
        returns number of transferred bytes
        (FEMAG 8.5 Rev 3282 or greater only)
        """
        ret = []
        total = 0
        chunk_size = 20*1024
        fnames = files
        if isinstance(files, str):
            fnames = [files]
        for fn in fnames:
            self.request_socket.send_string('CONTROL', flags=zmq.SNDMORE)
            basename = os.path.basename(fn)
            logger.info("upload %s --> %s", fn, basename)
            self.request_socket.send_string('upload = {}'.format(basename),
                                            flags=zmq.SNDMORE)
            with open(fn, mode="rb") as file:
                while True:
                    data = file.read(chunk_size)
                    if not data:
                        break
                    more = 0 if len(data) < chunk_size else zmq.SNDMORE
                    self.request_socket.send(data, flags=more)
                    total += len(data)
            ret += [s.decode('latin1')
                    for s in self.request_socket.recv_multipart()]
        return ret

    def copyfile(self, filename, dirname):
        """copy filename to dirname (FEMAG 9.2)"""
        if not self.request_socket:
            self.request_socket = self.__req_socket()
        self.request_socket.send_string('CONTROL', flags=zmq.SNDMORE)
        self.request_socket.send_string(f'copyfile {filename} {dirname}')
        return [r.decode() for r in self.request_socket.recv_multipart()]

    def change_case(self, dirname):
        """change case to dirname (FEMAG 9.2)"""
        logger.info("change_case to :  %s", dirname)
        if not self.request_socket:
            self.request_socket = self.__req_socket()
        self.request_socket.send_string('CONTROL', flags=zmq.SNDMORE)
        self.request_socket.send_string(f'casedir {dirname}')
        return [r.decode() for r in self.request_socket.recv_multipart()]

    def delete_case(self, dirname):
        """delete case dir (FEMAG 9.2)"""
        if not self.request_socket:
            self.request_socket = self.__req_socket()
        self.request_socket.send_string('CONTROL', flags=zmq.SNDMORE)
        self.request_socket.send_string(f'casedir -{dirname}')
        return [r.decode() for r in self.request_socket.recv_multipart()]

    def clear(self, timeout=2000):
        """clear lua script session"""
        return [r.decode('latin1')
                for r in self.send_request(['CONTROL', 'clear'],
                                           timeout=timeout)]

    def cleanup(self, timeout=2000):
        """remove all FEMAG files in working directory
        (FEMAG 8.5 Rev 3282 or greater only)"""
        return [r.decode('latin1')
                for r in self.send_request(['CONTROL', 'cleanup'],
                                           timeout=timeout)]

    def release(self):
        """signal finish calculation task to load balancer to free resources
        (Docker Cloud environment only)
        """
        return [r.decode('latin1')
                for r in self.send_request(['close'], timeout=1000)]

    def info(self, timeout=2000):
        """get various resource information
        (FEMAG 8.5 Rev 3282 or greater only)"""
        response = [r.decode('latin1')
                    for r in self.send_request(['CONTROL', 'info'],
                                               timeout=timeout)]
        status = json.loads(response[0])['status']
        if status == 'ok':
            try:
                self.femaghost = json.loads(response[1])['addr']
                logger.info("Set femaghost %s", self.femaghost)
            except KeyError:
                # ignore femaghost setting if no addr is returned (Windows only)
                pass
        return response

    def publishLevel(self, level):
        """set publish level"""
        return [r.decode('latin1')
                for r in self.send_request(['CONTROL', 'publish = {}'
                                            .format(level)], timeout=10000)]

    def getfile(self, filename=''):
        """get file (FEMAG 8.5 Rev 3282 or greater only)"""
        response = self.send_request(
            ['CONTROL', f'getfile = {filename}'], timeout=1000)
        return [response[0].decode('latin1'),
                response[1] if len(response) > 1 else b'']

    def exportsvg(self, fslcmds, timeout=10000):
        """get svg format from fsl commands (if any graphic created)
        (since FEMAG 8.5 Rev 3343) """
        response = self.send_request(['SVG', fslcmds], timeout=timeout)
        try:
            rc = json.loads(response[0].decode('latin1'))
            if rc['status'] == 'ok':
                return self.getfile(rc['result_file'][0])
        except json.decoder.JSONDecodeError:
            logger.warning(response[0])
        return [s.decode('latin1') for s in response]

    def exportmesh(self, fslcmds, timeout=120000):
        """get svg format (with mesh) from fsl commands (if any graphic created)
        (since FEMAG v2024.1-17) """
        response = self.send_request(['SVG+Mesh', fslcmds], timeout=timeout)
        try:
            rc = json.loads(response[0].decode('latin1'))
            if rc['status'] == 'ok':
                return self.getfile(rc['result_file'][0])
        except json.decoder.JSONDecodeError:
            logger.warning(response[0])
        return [s.decode('latin1') for s in response]

    def airgap_flux_density(self, pmod):
        # try to read bag.dat
        agr = self.getfile("bag.dat")
        status = json.loads(agr[0])['status']
        if status == 'ok':
            datfile = os.path.join(self.workdir, 'bag.dat')
            with open(datfile, 'wb') as bagfile:
                bagfile.write(agr[1])
            agi = ag.read(datfile, pmod)
        else:
            import numpy as np
            # try to read model file (TODO download with getfile)
            nc = self.read_nc()
            ag_elmnts = nc.airgap_center_elements
            logger.info("Airgap elements %d scale_factor %f",
                        len(ag_elmnts), nc.scale_factor())
            if len(ag_elmnts) < 1:
                raise ValueError("Missing airgap elements")
            scf = 360/nc.scale_factor()/ag_elmnts[-1].center[0]
            pos = np.array([e.center[0]*scf for e in ag_elmnts])
            bxy = np.array([e.flux_density() for e in ag_elmnts]).T
            if np.max(bxy[0]) > np.max(bxy[1]):
                agi = ag.fft(pos, bxy[0], pmod)
            else:
                agi = ag.fft(pos, bxy[1], pmod)
        return dict(Bamp=agi['Bamp'],
                    phi0=agi['phi0'],
                    angle=agi['pos'],
                    angle_fft=agi['pos'],
                    B=agi['B'],
                    B_fft=agi['B_fft'])

    def interrupt(self):
        """send push message to control port to stop current calculation"""
        context = zmq.Context.instance()
        if not self.femaghost:
            self.femaghost = '127.0.0.1'
        ctrl = context.socket(zmq.PUSH)
        ctrl.connect('tcp://{0}:{1}'.format(
            self.femaghost, self.port+2))
        logger.info("Interrupt %s", self.femaghost)
        ctrl.send_string('interrupt')
        ctrl.close()
        femagtools.zmq.SubscriberTask.clear()

    def copy_winding_file(self, name, wdg):
        wdg.write(name, self.workdir)
        self.upload(os.path.join(self.workdir, name+'.WID'))

    def copy_magnetizing_curves(self, model, dir=None, recsin='', feloss=''):
        """extract mc names from model and write files into workdir or dir if given
           and upload to Femag

        Return:
            list of extracted mc names (:obj:`list`)
        """
        dest = dir if dir else self.workdir
        mc_names = [m for m in model.set_magcurves(
            self.magnetizingCurves, self.magnets)]
        if isinstance(feloss, (int, float)):
            try:
                feloss = {1: 'jordan', 11: 'bertotti'}[int(feloss)]
            except KeyError:
                feloss = ''
        for m in mc_names:
            f = self.magnetizingCurves.writefile(m[0], dest,
                                                 fillfac=m[1],
                                                 recsin=recsin,
                                                 feloss=feloss)
            self.upload(os.path.join(dest, f))
        return mc_names

    def __call__(self, machine, simulation):
        """setup fsl file, run calculation and return BCH results
        Args:
          machine: dict with machine parameters or name of model
          simulation; dict with simulation parameters

        Raises:
           FemagError
        """
        if isinstance(machine, str):
            modelpars = dict(name=machine)
        else:
            modelpars = machine
        if 'exit_on_end' not in modelpars:
            modelpars['exit_on_end'] = 'false'
        if 'exit_on_error' not in modelpars:
            modelpars['exit_on_error'] = 'false'
        response = self.send_fsl(['save_model("close")'] +
                                 self.create_fsl(modelpars,
                                                 simulation))
        r = json.loads(response[0])
        if r['status'] != 'ok':
            raise FemagError(r['message'])

        if not simulation:
            model = self.model.props()
            model['name'] = machine['name']
            return [r, model]

        result_file = r['result_file'][0]
        if simulation['calculationMode'] == "pm_sym_loss":
            return self.read_los(self.modelname)

        if simulation['calculationMode'] == "asyn_motor":
            results = self.read_asm(self.modelname)
            if 'airgap_induc' in simulation:
                try:
                    pmod = results['p_gen']
                except KeyError:
                    pmod = 0
                bagdat = os.path.join(self.workdir, 'bag.dat')
                results['airgap'] = ag.read(bagdat, pmod=pmod)
            return results

        if simulation['calculationMode'] == "calc_field_ts":
            return self.read_ts(self.modelname)

        status, content = self.getfile(result_file)
        r = json.loads(status)
        if r['status'] == 'ok':
            bch = femagtools.bch.Reader()
            bch.read(content.decode('latin1'))
            bch = self.readResult(simulation, bch)
            return bch
        raise FemagError(r['message'])



import numpy as np
import femagtools.machine
from femagtools.job import Task
import pathlib
import pytest


@pytest.fixture
def data_dir():
    return pathlib.Path(__file__).with_name('data')


def test_ecsim_eval(data_dir):
    task = Task(id=1, directory=data_dir / 'parident')
    eval = femagtools.machine.im._eval_ecsim()
    r = eval(task)
    expected = {'r2': 0.00011541,
                'ls2': 3.57991e-07,
                'zeta2': 0.9891,
                'pl2v': 0.913}
    assert expected['r2'] == r['r2']
    assert expected['ls2'] == r['ls2']
    assert expected['zeta2'] == round(r['zeta2'], 4)
    assert expected['pl2v'] == round(r['pl2v'], 3)


def test_noloadrot_eval(data_dir):
    task = Task(id=1, directory=data_dir / 'parident')
    i1tab = [2.0384197650236815, 4.076839530047363,
             6.115259295071045, 8.153679060094726, 10.192098825118407]
    u1ph = 400/np.sqrt(3)
    pmod = 0
    eval = femagtools.machine.im._eval_noloadrot(pmod)
    r = eval(task)
    i0 = [2.0384,  4.0768,  6.1153,  8.1537, 10.1921]
    psi1_0 = [0.2725, 0.5301, 0.7106, 0.8091, 0.8641]
    #Bamp = [0.3091, 0.6013, 0.8058, 0.9159, 0.9768]
    Bamp = [0.3001, 0.5836, 0.7809, 0.8866, 0.9453]
    assert i0 == [round(x, 4) for x in r['i1_0']]
    assert psi1_0 == [round(x, 4) for x in np.mean(r['psi1_0'], axis=1)]
    assert Bamp == [round(x, 4) for x in np.mean(r['Bamp'], axis=1)]



# -*- coding: utf-8 -*-
"""
    femagtools.grid
    ~~~~~~~~~~~~~~~

    Parameter range calculation using Grid sampling



"""
import logging
import femagtools.parstudy
import numpy as np

logger = logging.getLogger(__name__)


def create_parameter_range(domain):
    """returns the transposed array of the combined domain values"""
    L = [len(d) for d in domain]
    LS = np.prod(L)
    s = []
    e = 1
    for d in domain:
        LS = LS//len(d)
        s.append(np.repeat(d*LS, e))
        e = e*L[0]
        L = L[1:]
    return np.array(s).T


class Grid(femagtools.parstudy.ParameterStudy):
    """Grid Parameter variation calculation"""

    def __init__(self, workdir,
                 magnetizingCurves=None, magnets=None, condMat=[],
                 result_func=None):  # tasktype='Task'):
        super().__init__(workdir,
                         magnetizingCurves, magnets, condMat,
                         result_func)

    def _get_names_and_range(self, dvars, num_samples):
        if isinstance(dvars, dict):
            dvarnames = dvars['columns']
            par_range = dvars['list']
            domain = [r for r in zip(*par_range)]
        else:
            dvarnames = [d['name'] for d in dvars]
            domain = [list(np.linspace(d['bounds'][0],
                                       d['bounds'][1],
                                       d['steps'])) if 'bounds' in d else d['values']
                      for d in dvars]

            par_range = create_parameter_range(domain)
        return dvarnames, domain, par_range



"""manage multicore/multiprocessing jobs

"""
import platform
import multiprocessing
import subprocess
import time
import os
import re
import threading
import pathlib
import logging
from .job import Job
import femagtools.config as cfg
import femagtools.zmq
from femagtools.zmq import SubscriberTask
try:
    from subprocess import DEVNULL
except ImportError:
    DEVNULL = open(os.devnull, 'wb')

logger = logging.getLogger(__name__)

class LicenseError(Exception):
    pass

class ProgressLogger(threading.Thread):
    def __init__(self, dirs, num_cur_steps, timestep, notify):
        threading.Thread.__init__(self)
        self.protfiles = [femagtools.zmq.ProtFile(d, num_cur_steps)
                          for d in dirs]
        self.numTot = len(dirs)
        self.running = False
        self.timestep = timestep
        self.notify = notify

    def run(self):
        self.running = True
        while self.running:
            if self.timestep > 0:
                time.sleep(self.timestep)
            logmsg = [p.update() for p in self.protfiles]
            summary = [l  # f'<{i}> {l}'
                       for i, l in enumerate(logmsg)
                       if l]
            if summary:
                labels = set([p.name for p in self.protfiles])
                logger.info('%s: %s',
                            ', '.join(labels),
                            ', '.join(summary))
                if self.notify:
                    numOf = f"{summary.count('100.0%')} of {self.numTot}"
                    percent = sum([float(i[:-1])
                                   for i in summary]) / self.numTot
                    self.notify(
                        ["progress_logger",
                         f"{self.numTot}:{numOf}:{percent}:{' '.join(summary)}"])
            else:
                # TODO: log message might be misleading
                logger.debug('collecting FE losses ...')
                return

    def stop(self):
        self.running = False


def run_femag(cmd, workdir, fslfile, port):
    """Start the femag command as subprocess.

    :internal:

    Args:
        cmd: (list) The program (executable image) to be run
        workdir: The workdir where the calculation files are stored
        fslfile: The name of the start file (usually femag.fsl)
    """
    with open(os.path.join(workdir, "femag.out"), "wb") as out, \
            open(os.path.join(workdir, "femag.err"), "wb") as err:
        try:
            args = ['-b', str(port), fslfile] if port else ['-b', fslfile]
            proc = subprocess.Popen(cmd + args,
                                    shell=False,
                                    stdin=DEVNULL,
                                    stdout=out,
                                    stderr=err,
                                    cwd=workdir)
            logger.info('%s (pid %d, workdir %s)', cmd, proc.pid, workdir)
            # write pid file
            with open(os.path.join(workdir, 'femag.pid'), 'w') as pidfile:
                pidfile.write("{}\n".format(proc.pid))

            # wait
            proc.wait()
            os.remove(os.path.join(workdir, 'femag.pid'))

            #logger.info("Finished pid: %d return %d",
            #            proc.pid, proc.returncode)
            #return proc.returncode
        except OSError as e:
            logger.error("Starting process failed: %s, Command: %s", e, cmd)
            raise

    # raise License Error
    if proc.returncode != 0:
        with open(os.path.join(workdir, "femag.err"), "r") as err:
            for line in err:
                if 'license' in line:
                    raise LicenseError(line)

    logger.info("Finished pid: %d return %d", proc.pid, proc.returncode)
    return proc.returncode


class Engine:
    """The MultiProc engine uses a pool of local calculation processes.

    This is more or less a decorator for the `Python multiprocessing Module
    <https://docs.python.org/3.6/library/multiprocessing.html>`_

    Args:
        cmd: the program (executable image) to be run
            (femag dc is used if None)
        process_count: number of processes (cpu_count() if None)
        timestep: time step in seconds for progress log messages if > 0)
    """

    def __init__(self, **kwargs):
        self.process_count = kwargs.get('process_count', None)
        self.notify = kwargs.get('notify', None)
        # cogg_calc mode, subscribe xyplot
        self.calc_mode = kwargs.get('calc_mode')
        self.port = kwargs.get('port', 0)
        self.curve_label = kwargs.get('curve_label')
        cmd = kwargs.get('cmd', '')
        if cmd:
            self.cmd = [cmd]
            if platform.system() == 'Windows':
                self.cmd.append('-m')

        self.progressLogger = 0
        self.progress_timestep = kwargs.get('timestep', -1)
        self.subscriber = None
        self.job = None
        self.tasks = []

    def create_job(self, workdir):
        """Create a FEMAG :py:class:`Job`

        Args:
            workdir: The workdir where the calculation files are stored

        Return:
            FEMAG :py:class:`Job`
        """
        self.job = Job(workdir)
        return self.job

    def submit(self, extra_result_files=[]):
        """Starts the FEMAG calculation(s) with the internal
        :py:meth:`multiproc.run_femag` function

        Return:
            length of started tasks
        """
        # must check if cmd is set:
        args = []
        if platform.system() == 'Windows':
            args.append('-m')

        try:
            for t in self.job.tasks:
                t.cmd = self.cmd
        except AttributeError:
            for t in self.job.tasks:
                t.cmd = [cfg.get_executable(
                    t.stateofproblem)] + args

        num_proc = self.process_count
        if not num_proc and multiprocessing.cpu_count() > 1:
            num_proc = min(multiprocessing.cpu_count()-1, len(self.job.tasks))
        self.pool = multiprocessing.Pool(num_proc)
        if self.port:
            header = [b'progress']
            if self.calc_mode == 'cogg_calc':
                header +=[b'xyplot']
            self.subscriber = [SubscriberTask(port=self.port + i * 5,
                                              host='127.0.0.1',
                                              notify=self.notify,
                                              header=header,
                                              curve_label=self.curve_label,
                                              num_cur_steps=self.job.num_cur_steps,
                                              timestep=self.progress_timestep
                                              )
                               for i, t in enumerate(self.job.tasks)]
            [s.start() for s in self.subscriber]
            self.tasks = [self.pool.apply_async(
                run_femag, args=(t.cmd, t.directory, t.fsl_file, self.port + i * 5))
                          for i, t in enumerate(self.job.tasks)]
        else:
            self.tasks = [self.pool.apply_async(
                run_femag, args=(t.cmd, t.directory, t.fsl_file, 0))
                          for t in self.job.tasks]
        self.pool.close()

        # only works on linux
        if (self.progress_timestep and not self.port and
            self.job.num_cur_steps):
            self.progressLogger = ProgressLogger(
                [t.directory for t in self.job.tasks],
                num_cur_steps=self.job.num_cur_steps,
                timestep=self.progress_timestep,
                notify=self.notify)
            self.progressLogger.start()
        return len(self.tasks)

    def join(self):
        """Wait until all calculations are finished

        Return:
            list of all calculations status (C = Ok, X = error)
        """
        exitcodes = [task.get() for task in self.tasks]
        status = []
        for t, ec in zip(self.job.tasks, exitcodes):
            t.status = 'C'
            if ec != 0:
                t.status = 'X'
                errmsg = pathlib.Path(t.directory) / 'femag.err'
                if errmsg.exists():
                    t.errmsg = errmsg.read_text()
                    if t.errmsg:
                        logger.error(t.errmsg)
            status.append(t.status)
        self.stopThreads()
        self.pool = None # garbage collector deletes threads
        return status

    def read_nc(self):
        """return a generator object of nc list"""
        for t in self.job.tasks:
            yield t.read_nc()
        return None

    def stopThreads(self):
        """ stop all running treads
        """
        if self.progressLogger:
            self.progressLogger.stop()
        if self.port and self.subscriber:
            [s.stop() for s in self.subscriber]
            SubscriberTask.clear()
            self.subscriber = None

    def terminate(self):
        """ terminate all
        """
        logger.info("terminate Engine")
        self.stopThreads()

        # terminate pool
        try:
            if self.pool:
                self.pool.terminate()
                self.pool = None # garbage collector deletes threads
        except AttributeError as e:
            logger.warn("%s", e)



# -*- coding: utf-8 -*-
"""
  femagtools.dxfsl.concat
  ~~~~~~~~~~~~~~~~~~~~~~~

  Authors: Ronald Tanner, Beat Holm
"""
from __future__ import print_function

import numpy as np
import logging
import sys
import json
import time
from pathlib import Path

logger = logging.getLogger('femagtools.journal')
journal = None

#############################
#          journal          #
#############################


def getJournal(name=None, aktiv=False):
    global journal
    if not journal:
        if not name:
            name = 'none'
        journal = Journal(name=name, aktiv=aktiv)
    return journal


class Journal(object):
    def __init__(self,
                 name="journal",
                 aktiv=False):
        self.name = name
        self.aktiv = aktiv
        self.journal = {}
        self.data = {}
        self.data_key = None
        self.filename = Path('{}.json'.format(self.name))

    def open_journal(self):
        if self.filename.is_file():
            with open(self.filename, 'r') as f:
                try:
                    self.journal = json.loads(f.read())
                except:
                    self.journal = {}
                f.close()

    def write_journal(self):
        if not self.journal:
            return
        self.set_benchmark()
        self.sort_by_keys()
        with open(self.filename, 'w') as f:
            f.write(json.dumps(self.journal, indent=4))
            f.close()

    def sort_by_keys(self):
        myKeys = list(self.data.keys())
        myKeys.sort()
        sorted_dict = {i: self.data[i] for i in myKeys}
        self.data = sorted_dict
        self.journal[self.data_key] = sorted_dict

    def get_journal(self, name):
        if not self.aktiv:
            return None

        if self.data:
            return self.data

        self.open_journal()
        self.data = self.journal.get(name, None)
        self.data = {'filename': ""}  # initialise
        self.data_key = name
        self.journal[name] = self.data
        return self.data

    def get_total(self, name):
        val = self.data.get(name, None)
        if val is None:
            return 0
        if isinstance(val, list):
            if not val:
                return 0
            if isinstance(val[0], int) or isinstance(val[0], float):
                val = sum(val)
                return val
            return len(val)
        if isinstance(val, int) or isinstance(val, float):
            return val
        return 1  # entries

    def set_benchmark(self):
        if self.get_total('area_errors') > 0:
            self.put_warning("Problem with areas")
        if self.get_total('appendices_deleted') > 0:
            self.put_warning("Problem with appendices")

    def set(self, name, val):
        if not self.data:
            return
        self.data[name] = val

    def put(self, name, val):
        if not self.data:
            return
        data = self.data.get(name, None)
        if data is None:
            self.data[name] = val
            return
        if isinstance(data, list):
            self.data[name].append(val)
            return
        data_list = [data]
        data_list.append(val)
        self.data[name] = data_list

    def set_filename(self, val):
        self.set('filename', val)

    def put_areas(self, val):
        self.put('areas', val)

    def put_area_errors(self, val):
        self.put('area_errors', val)

    def put_elements(self, val):
        self.put('elements', val)

    def put_nodes(self, val):
        self.put('nodes', val)

    def put_concat_lines(self, val):
        self.put('concat_lines', val)

    def put_concat_arcs(self, val):
        self.put('concat_arcs', val)

    def put_appendices(self, val):
        self.put('appendices', val)

    def put_appendices_connected(self, val):
        self.put('appendices_connected', val)

    def put_appendices_remaining(self, val):
        self.put('appendices_remaining', val)

    def put_appendices_deleted(self, val):
        self.put('appendices_deleted', val)

    def put_nodes_connected(self, val):
        self.put('nodes_connected', val)

    def put_exception(self, msg):
        self.put('exception', msg)

    def put_warning(self, msg):
        self.put('warning', msg)



import femagtools.hxy
import pathlib
import pytest


@pytest.fixture
def data_dir():
    return pathlib.Path(__file__).with_name('data') / 'hxy'


def test_read(data_dir):
    num_magnets = 2
    magnets = femagtools.hxy.read(data_dir / 'PM270L8_011.hxy',
                                  num_magnets)
    assert len(magnets) == num_magnets
    assert set([len(m['e'][0]) for m in magnets]) == {169, 189}
    assert [m['pos'][0] for m in magnets] == [0.0, 0.0]
    assert sorted([m['havg'][0] for m in magnets]) == pytest.approx([156.3, 156.5], 0.1)
    assert sorted([m['hmax'][0] for m in magnets]) == pytest.approx([195.6, 304.4], 0.1)



# -*- coding: utf-8 -*-
"""Fitting methods for loss coeffs


"""
import numpy as np
import scipy.optimize as so
import scipy.interpolate as ip
import logging


def pfe_bertotti(f, B, ch, alpha, cw, ce):
    return ch*f*B**alpha + cw*f**2*B**2 + ce*f**1.5*B**1.5


def pfe_jordan(f, B, ch, fh, cw, fw, fb, fo, Bo):
    return (ch*(f/fo)**fh + cw*(f/fo)**fw)*(B/Bo)**fb


def pfe_steinmetz(f, B, cw, fw, fb, fo, Bo):
    return cw*(f/fo)**fw * (B/Bo)**fb

def pfe_modified_steinmetz(f, B, ch, cw, alpha, beta, fo, Bo):
    return ch*(f/fo)*(B/Bo)**(alpha + B*beta) + cw*(f/fo)**2*(B/Bo)**2

def wbert(f, b, ch, cw, cx):
    return (ch + cw*f)*b**2 + cx*f**0.5*b**1.5

def fitsteinmetz(f, B, losses, Bo, fo, alpha0=1.0):
    """fit coeffs of
    losses(f,B)=cw*(f/fo)**alfa*(B/Bo)**beta
    returns (cw, alfa, beta)
    """
    if np.isscalar(f):
        fbx = [[f]*len(losses), B]
        y = losses
        fitp, cov = so.curve_fit(
            lambda x, cw, beta: pfe_steinmetz(
                x[0], x[1], cw, alpha0, beta, fo, Bo),
            fbx, y, (1.0, 2.0))
        fitp = np.insert(fitp, 1, alpha0)

    else:
        pfe = losses
        z = []
        for i, fx in enumerate(f):
            if fx:
                if isinstance(B[0], float):
                    z += [(fx, bx, y)
                          for bx, y in zip(B, pfe[i])
                          if isinstance(y, float)]
                else:
                    z += [(fx, bx, y)
                          for bx, y in zip(B[i], pfe[i]) if y]

        fbx = np.array(z).T[0:2]
        y = np.array(z).T[2]
        fitp, cov = so.curve_fit(
            lambda x, cw, alpha, beta: pfe_steinmetz(
                x[0], x[1], cw, alpha, beta, fo, Bo),
            fbx, y, (1.0, 1.0, 2.0))
    return fitp

def fit_modified_steinmetz(f, B, losses, Bo, fo):
    """fit coeffs of modifeld steinmetz
    losses(f,B)=ch*(f/fo)*(B/Bo)**(alpha + B*beta) + cw*(f/fo)**2*(B/Bo)**2
    returns (ch, cw, alpha, beta)
    """
    pfe = losses
    z = []
    for i, fx in enumerate(f):
        if fx:
            if isinstance(B[0], float):
                z += [(fx, bx, y)
                      for bx, y in zip(B, pfe[i])
                      if y]
            else:
                z += [(fx, bx, y)
                      for bx, y in zip(B[i], pfe[i])
                      if y]

    fbx = np.array(z).T[0:2]
    y = np.array(z).T[2]
    fitp, cov = so.curve_fit(lambda x, ch, cw, alpha, beta: pfe_modified_steinmetz(
        x[0], x[1], ch, cw, alpha, beta, fo, Bo),
        fbx, y, (1.0, 1.0, 1.0, 1.0))
    return fitp

def fitjordan(f, B, losses, Bo, fo):
    """fit coeffs of
    losses(f,B)=(ch*(f/fo)**alpha + ch*(f/fo)**beta)*(B/Bo)**gamma
    returns (ch, alpha, cw, beta, gamma)
    """
    pfe = losses
    z = []
    for i, fx in enumerate(f):
        if fx:
            if isinstance(B[0], float):
                z += [(fx, bx, y)
                      for bx, y in zip(B, pfe[i])
                      if y]
            else:
                z += [(fx, bx, y)
                      for bx, y in zip(B[i], pfe[i])
                      if y]

    fbx = np.array(z).T[0:2]
    y = np.array(z).T[2]
    fitp, cov = so.curve_fit(lambda x, ch, alpha, cw, beta, gamma: pfe_jordan(
        x[0], x[1], ch, alpha, cw, beta, gamma, fo, Bo),
        fbx, y, (1.0, 1.0, 1.0, 2.0, 1.0))
    return fitp


def fit_bertotti(f, B, losses):
    """fit coeffs of
    losses(f,B)=(ch*f + ch*f**2)*B**2 + ce*f**1.5*B**1.5
    returns (ch, cw, ce)

    Args:
        f: list of n strictly increasing frequency values
        B: list of n or list of list of induction values (nxm)
        losses: list of list of fe loss values (nxm)
    """
    i0 = 0
    if np.isclose(f[i0], 0):
        i0 = 1
    v = []
    # collect losses and flux density along the frequency
    for k in range(len(losses[i0])):
        y = []
        bb = []
        if isinstance(B[0], tuple) or isinstance(B[0], list):
            for fx, bx, p in zip(f[i0:], B[i0:], losses[i0:]):
                if k < len(p):
                    y.append(p[k]/fx)
                    bb.append(bx[k])
        else:
            for fx, p in zip(f[i0:], losses[i0:]):
                if k < len(p):
                    y.append(p[k]/fx)
                    bb.append(B[k])
        j = len(y)
        if j > 2:
            v.append(np.array((f[i0:j+i0], bb, y)).T.tolist())

    z = np.array([b for a in v for b in a]).T
    fbx = z[0:2]
    y = z[2]
    fitp, _ = so.curve_fit(
        lambda x, ch, cw, ce: wbert(
            x[0], x[1], ch, cw, ce),
        fbx, y,
        bounds=[(0, 0, 0),
                (np.inf, np.inf, np.inf)])
    return fitp

def fit_bertotti0(f, B, losses, generate=False):
    """fit coeffs of
    losses(f,B)=(ch*f + ch*f**2)*B**2 + ce*f**1.5*B**1.5
    returns (ch, cw, ce)

    Args:
        f: list of n strictly increasing frequency values
        B: list of list of induction values (nxm)
        losses: list of list of fe loss values (nxm)
        generate: generates additional frequency samples if True
    """
    pb = [ip.CubicSpline(bi, pi)
          for bi, pi in zip(B, losses)]
    i0 = 0
    if np.isclose(f[i0], 0):
        i0 = 1
    wy = [pb[i+i0](1.0)/f[i+i0]
          for i in range(len(f[i0:]))]
    csf = ip.CubicSpline(np.sqrt(f[i0:]), wy)
    xx = np.linspace(0, np.sqrt(f[-1]), 10)
    z = np.polyfit(xx, csf(xx), 2)
    ch0 = z[2]
    cw0 = z[0]
    ce0 = z[1]

    v = []
    df = 70
    # collect losses and flux density along the frequency
    for k in range(len(losses[i0])):
        y = []
        bb = []
        for fx, bx, p in zip(f[i0:], B[i0:], losses[i0:]):
            if k < len(p):
                y.append(p[k]/fx)
                bb.append(bx[k])
            else:
                break
        j = len(y)
        if j > 2:
            if generate:
                # generate additional samples to improve LM fit (experimental)
                nsteps = int(np.ceil((f[i0:][j-1] - f[i0])/df))
                fw = ip.CubicSpline(f[i0:j+1], y)
                bw = ip.CubicSpline(f[i0:j+1], bb)
                fx = np.linspace(f[i0], f[i0:][j-1], nsteps)
                v.append(np.array((fx, bw(fx), fw(fx))).T)
            else:
                v.append(np.array((f[i0:j+i0], bb, y)).T.tolist())


    def wbert(f, b, ch, cw, cx):
        return (ch + cw*f)*b**2 + cx*f**0.5*b**1.5


    z = np.array([b for a in v for b in a]).T
    fbx = z[0:2]
    y = z[2]
    guess = (ch0, cw0, ce0)
    fitp, _ = so.curve_fit(
        lambda x, ch, cw, ce: wbert(
            x[0], x[1], ch, cw, ce),
        fbx, y,
        bounds=[(0, 0, 0),
                (np.inf, np.inf, np.inf)])
    return fitp


def fit_bertotti1(f, B, losses):
    """fit coeffs of
    losses(f,B)=ch*f*B**alpha + ch*f**2*B**2 + ce*f**1.5*B**1.5
    returns (ch, alpha, cw, ce)
    """
    v = []
    i0 = 0
    if np.isclose(f[i0], 0):
        i0 = 1
    df = 70  # delta frequency
    # collect losses and flux density along the frequency
    for k in range(len(losses[i0])):
        y = []
        bb = []
        for fx, bx, p in zip(f[i0:], B[i0:], losses[i0:]):
            if k < len(p):
                y.append(p[k]/fx)
                bb.append(bx[k])
            else:
                break
        j = len(y)
        if j > 2:
            v.append(np.array((f[i0:j+1], bb, y)).T)

    def wbert(f, b, ch, alpha, cw, cx):
        return ch*b**alpha + cw*f*b**2 + cx*f**0.5*b**1.5

    z = np.array([b for a in v for b in a]).T
    fbx = z[0:2]
    y = z[2]

    fitp, cov = so.curve_fit(
            lambda x, ch, alpha, cw, cx: wbert(
                x[0], x[1], ch, alpha, cw, cx),
            fbx, y, (1e-3, 2, 1e-3, 1e-3))
    return fitp
